# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""This module contains FreeCAD commands for the BIM workbench"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate


class BIM_Reextrude:
    def GetResources(self):
        return {
            "Pixmap": "BIM_Reextrude",
            "MenuText": QT_TRANSLATE_NOOP("BIM_Reextrude", "Re-Extrude"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_Reextrude", "Recreates an extruded structure from a selected face"
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):
        import Draft
        import Part
        import Arch

        # get selected object and face
        sel = FreeCADGui.Selection.getSelectionEx()

        if (
            (len(sel) == 1)
            and (len(sel[0].SubObjects) == 1)
            and ("Face" in sel[0].SubElementNames[0])
        ):
            sel = sel[0]
            obj = sel.Object
            name = obj.Name
            label = obj.Label

            fac = sel.SubObjects[0]

            # make this undoable
            FreeCAD.ActiveDocument.openTransaction("Reextrude")

            # check if the face has holes or any of the edges is not a line
            wirable = True
            if len(fac.Wires) > 1:
                wirable = False
            else:
                for edge in fac.Edges:
                    if not isinstance(edge, (Part.Line, Part.LineSegment)):
                        # edge can be a spline, but even so be straight. Simple check if tangents are identical at first and last verts...
                        if (
                            edge.tangentAt(edge.FirstParameter).getAngle(
                                edge.tangentAt(edge.LastParameter)
                            )
                            > 0.0001
                        ):
                            wirable = False
                            break
            if wirable:
                # recompose the base wire
                verts = [v.Point for v in fac.Wires[0].OrderedVertexes]
                wir = Draft.makeWire(verts, closed=True)
            else:
                # there are curves. Unable to make a wire. We just use the base face
                wir = FreeCAD.ActiveDocument.addObject("Part::Feature", "Face")
                wir.Shape = fac

            # make the new object
            if Draft.getType(obj) == "Wall":
                newobj = Arch.makeWall(wir)
            elif Draft.getType(obj) == "Panel":
                newobj = Arch.makePanel(wir)
            else:
                newobj = Arch.makeStructure(wir)

            # deduce the normal and extrusion size
            norm = fac.normalAt(0, 0).negative()
            newobj.Normal = norm
            for e in obj.Shape.Edges:
                if abs(e.tangentAt(0).getAngle(norm)) < 0.0001:
                    if hasattr(newobj, "Thickness"):
                        newobj.Thickness = e.Length
                    else:
                        newobj.Height = e.Length

            # set material
            if hasattr(obj, "Material") and obj.Material:
                newobj.Material = obj.Material

            # set role and class
            if hasattr(obj, "IfcType"):
                newobj.IfcType = obj.IfcType
            elif hasattr(obj, "IfcRole"):
                newobj.IfcRole = obj.IfcRole
            if hasattr(obj, "StandardCode"):
                newobj.StandardCode = obj.StandardCode

            # update objects relating to this one
            for parent in obj.InList:
                for prop in parent.PropertiesList:
                    if getattr(parent, prop) == obj:
                        setattr(parent, prop, newobj)
                        FreeCAD.Console.PrintMessage(
                            "Object "
                            + parent.Label
                            + "'s reference to this object has been updated\n"
                        )
                    elif isinstance(getattr(parent, prop), list) and (obj in getattr(parent, prop)):
                        if (prop == "Group") and hasattr(parent, "addObject"):
                            parent.addObject(newobj)
                        else:
                            g = getattr(parent, prop)
                            g.append(newobj)
                            setattr(parent, prop, g)
                        FreeCAD.Console.PrintMessage(
                            "Object "
                            + parent.Label
                            + "'s reference to this object has been updated\n"
                        )
                    # TODO treat PropertyLinkSub / PropertyLinkSubList

            # delete original object
            FreeCAD.ActiveDocument.removeObject(name)
            newobj.Label = label

            # commit changes
            FreeCAD.ActiveDocument.commitTransaction()
            FreeCAD.ActiveDocument.recompute()

        else:
            FreeCAD.Console.PrintError(
                translate("BIM", "Error: Select exactly one base face") + "\n"
            )


FreeCADGui.addCommand("BIM_Reextrude", BIM_Reextrude())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from PySide import QtCore, QtGui
from PySide.QtCore import QT_TRANSLATE_NOOP
from pivy import coin
import FreeCAD
import FreeCADGui
import Path
import Path.Base.Gui.GetPoint as PathGetPoint
import Path.Dressup.Tags as PathDressupTag
import PathScripts.PathUtils as PathUtils
import Path.Dressup.Utils as PathDressup


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate


def addDebugDisplay():
    return Path.Log.getLevel(Path.Log.thisModule()) == Path.Log.Level.DEBUG


class PathDressupTagTaskPanel:
    DataX = QtCore.Qt.ItemDataRole.UserRole
    DataY = QtCore.Qt.ItemDataRole.UserRole + 1
    DataZ = QtCore.Qt.ItemDataRole.UserRole + 2
    DataID = QtCore.Qt.ItemDataRole.UserRole + 3

    def __init__(self, obj, viewProvider, jvoVisibility=None):
        self.obj = obj
        self.obj.Proxy.obj = obj
        self.viewProvider = viewProvider
        self.form = FreeCADGui.PySideUic.loadUi(":/panels/HoldingTagsEdit.ui")
        self.getPoint = PathGetPoint.TaskPanel(self.form.removeEditAddGroup, True)
        self.jvo = PathUtils.findParentJob(obj).ViewObject
        if jvoVisibility is None:
            FreeCAD.ActiveDocument.openTransaction("Edit HoldingTags Dress-up")
            self.jvoVisible = self.jvo.isVisible()
            if self.jvoVisible:
                self.jvo.hide()
                self.obj.ViewObject.show()
        else:
            self.jvoVisible = jvoVisibility
        self.pt = FreeCAD.Vector(0, 0, 0)

        self.isDirty = True
        self.buttonBox = None
        self.tags = None
        self.Positions = None
        self.Disabled = None
        self.editItem = None

    def getStandardButtons(self):
        return (
            QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Apply | QtGui.QDialogButtonBox.Cancel
        )

    def clicked(self, button):
        if button == QtGui.QDialogButtonBox.Apply:
            self.getFields()
            self.obj.Proxy.execute(self.obj)
            self.isDirty = False

    def modifyStandardButtons(self, buttonBox):
        self.buttonBox = buttonBox
        self.getPoint.buttonBox = buttonBox

    def abort(self):
        FreeCAD.ActiveDocument.abortTransaction()
        self.cleanup(False)

    def reject(self):
        FreeCAD.ActiveDocument.abortTransaction()
        self.cleanup(True)

    def accept(self):
        FreeCAD.ActiveDocument.commitTransaction()
        self.cleanup(True)
        if self.isDirty:
            self.getFields()
            FreeCAD.ActiveDocument.recompute()

    def cleanup(self, gui):
        self.viewProvider.clearTaskPanel()
        if gui:
            FreeCADGui.ActiveDocument.resetEdit()
            FreeCADGui.Control.closeDialog()
            FreeCAD.ActiveDocument.recompute()
            if self.jvoVisible:
                self.jvo.show()

    def getTags(self, includeCurrent):
        tags = []
        index = self.form.lwTags.currentRow()
        for i in range(0, self.form.lwTags.count()):
            item = self.form.lwTags.item(i)
            enabled = item.checkState() == QtCore.Qt.CheckState.Checked
            x = item.data(self.DataX)
            y = item.data(self.DataY)
            # print("(%.2f, %.2f) i=%d/%s" % (x, y, i, index))
            if includeCurrent or i != index:
                tags.append((x, y, enabled))
        return tags

    def getFields(self):
        width = FreeCAD.Units.Quantity(self.form.ifWidth.text()).Value
        height = FreeCAD.Units.Quantity(self.form.ifHeight.text()).Value
        angle = self.form.dsbAngle.value()
        radius = FreeCAD.Units.Quantity(self.form.ifRadius.text()).Value

        tags = self.getTags(True)
        positions = []
        disabled = []
        for i, (x, y, enabled) in enumerate(tags):
            positions.append(FreeCAD.Vector(x, y, 0))
            if not enabled:
                disabled.append(i)

        if width != self.obj.Width:
            self.obj.Width = width
            self.isDirty = True
        if height != self.obj.Height:
            self.obj.Height = height
            self.isDirty = True
        if angle != self.obj.Angle:
            self.obj.Angle = angle
            self.isDirty = True
        if radius != self.obj.Radius:
            self.obj.Radius = radius
            self.isDirty = True
        if positions != self.obj.Positions:
            self.obj.Positions = positions
            self.isDirty = True
        if disabled != self.obj.Disabled:
            self.obj.Disabled = disabled
            self.isDirty = True

    def updateTagsView(self):
        Path.Log.track()
        self.form.lwTags.blockSignals(True)
        self.form.lwTags.clear()
        for i, pos in enumerate(self.Positions):
            lbl = "%d: (%.2f, %.2f)" % (i, pos.x, pos.y)
            item = QtGui.QListWidgetItem(lbl)
            item.setData(self.DataX, pos.x)
            item.setData(self.DataY, pos.y)
            item.setData(self.DataZ, pos.z)
            item.setData(self.DataID, i)
            if i in self.Disabled:
                item.setCheckState(QtCore.Qt.CheckState.Unchecked)
            else:
                item.setCheckState(QtCore.Qt.CheckState.Checked)
            flags = QtCore.Qt.ItemFlag.ItemIsSelectable
            flags |= QtCore.Qt.ItemFlag.ItemIsEnabled
            flags |= QtCore.Qt.ItemFlag.ItemIsUserCheckable
            item.setFlags(flags)
            self.form.lwTags.addItem(item)
        self.form.lwTags.blockSignals(False)
        self.whenTagSelectionChanged()
        self.viewProvider.updatePositions(self.Positions, self.Disabled)

    def generateNewTags(self):
        count = self.form.sbCount.value()
        Path.Log.track(count)
        if not self.obj.Proxy.generateTags(self.obj, count):
            self.obj.Proxy.execute(self.obj)
        self.Positions = self.obj.Positions
        self.Disabled = self.obj.Disabled
        self.updateTagsView()

    def copyNewTags(self):
        sel = self.form.cbSource.currentText()
        tags = [o for o in FreeCAD.ActiveDocument.Objects if sel == o.Label]
        if 1 == len(tags):
            if not self.obj.Proxy.copyTags(self.obj, tags[0]):
                self.obj.Proxy.execute(self.obj)
            self.Positions = self.obj.Positions
            self.Disabled = self.obj.Disabled
            self.updateTagsView()
        else:
            Path.Log.error("Cannot copy tags - internal error")

    def updateModel(self):
        self.getFields()
        self.updateTagsView()
        self.isDirty = True

    def whenCountChanged(self):
        count = self.form.sbCount.value()
        self.form.pbGenerate.setEnabled(count)

    def selectTagWithId(self, index):
        Path.Log.track(index)
        self.form.lwTags.setCurrentRow(index)

    def whenTagSelectionChanged(self):
        index = self.form.lwTags.currentRow()
        count = self.form.lwTags.count()
        self.form.pbDelete.setEnabled(index != -1 and count > 2)
        self.form.pbEdit.setEnabled(index != -1)
        self.viewProvider.selectTag(index)

    def whenTagsViewChanged(self):
        self.updateTagsViewWith(self.getTags(True))

    def updateTagsViewWith(self, tags):
        self.tags = tags
        self.Positions = [FreeCAD.Vector(t[0], t[1], 0) for t in tags]
        self.Disabled = [i for (i, t) in enumerate(self.tags) if not t[2]]
        self.updateTagsView()

    def deleteSelectedTag(self):
        self.updateTagsViewWith(self.getTags(False))

    def addNewTagAt(self, point, obj):
        if point and obj and self.obj.Proxy.pointIsOnPath(self.obj, point):
            Path.Log.info("addNewTagAt(%.2f, %.2f)" % (point.x, point.y))
            self.Positions.append(FreeCAD.Vector(point.x, point.y, 0))
            self.updateTagsView()
        else:
            Path.Log.notice("ignore new tag at %s (obj=%s, on-path=%d" % (point, obj, 0))

    def addNewTag(self):
        self.tags = self.getTags(True)
        self.getPoint.getPoint(self.addNewTagAt)

    def editTagAt(self, point, obj):
        Path.Log.track(point, obj)
        if point and self.obj.Proxy.pointIsOnPath(self.obj, point):
            tags = []
            for i, (x, y, enabled) in enumerate(self.tags):
                if i == self.editItem:
                    tags.append((point.x, point.y, enabled))
                else:
                    tags.append((x, y, enabled))
            self.updateTagsViewWith(tags)

    def editTag(self, item):
        if item:
            self.tags = self.getTags(True)
            self.editItem = item.data(self.DataID)
            x = item.data(self.DataX)
            y = item.data(self.DataY)
            z = item.data(self.DataZ)
            self.getPoint.getPoint(self.editTagAt, FreeCAD.Vector(x, y, z))

    def editSelectedTag(self):
        self.editTag(self.form.lwTags.currentItem())

    def setFields(self):
        self.updateTagsView()
        self.form.sbCount.setValue(len(self.Positions))
        self.form.ifHeight.setText(
            FreeCAD.Units.Quantity(self.obj.Height, FreeCAD.Units.Length).UserString
        )
        self.form.ifWidth.setText(
            FreeCAD.Units.Quantity(self.obj.Width, FreeCAD.Units.Length).UserString
        )
        self.form.dsbAngle.setValue(self.obj.Angle)
        self.form.ifRadius.setText(
            FreeCAD.Units.Quantity(self.obj.Radius, FreeCAD.Units.Length).UserString
        )

    def setupUi(self):
        self.Positions = self.obj.Positions
        self.Disabled = self.obj.Disabled

        self.setFields()
        self.whenCountChanged()

        if self.obj.Proxy.supportsTagGeneration(self.obj):
            self.form.sbCount.valueChanged.connect(self.whenCountChanged)
            self.form.pbGenerate.clicked.connect(self.generateNewTags)
        else:
            self.form.cbTagGeneration.setEnabled(False)

        enableCopy = False
        for tags in sorted(
            [o.Label for o in FreeCAD.ActiveDocument.Objects if "DressupTag" in o.Name]
        ):
            if tags != self.obj.Label:
                enableCopy = True
                self.form.cbSource.addItem(tags)
        if enableCopy:
            self.form.gbCopy.setEnabled(True)
            self.form.pbCopy.clicked.connect(self.copyNewTags)

        self.form.lwTags.itemChanged.connect(self.whenTagsViewChanged)
        self.form.lwTags.itemSelectionChanged.connect(self.whenTagSelectionChanged)
        self.form.lwTags.itemActivated.connect(self.editTag)

        self.form.pbDelete.clicked.connect(self.deleteSelectedTag)
        self.form.pbEdit.clicked.connect(self.editSelectedTag)
        self.form.pbAdd.clicked.connect(self.addNewTag)

        self.viewProvider.turnMarkerDisplayOn(True)


class HoldingTagMarker:
    def __init__(self, point, colors):
        self.point = point
        self.color = colors
        self.sep = coin.SoSeparator()
        self.pos = coin.SoTranslation()
        self.pos.translation = (point.x, point.y, point.z)
        self.sphere = coin.SoSphere()
        self.scale = coin.SoType.fromName("SoShapeScale").createInstance()
        self.scale.setPart("shape", self.sphere)
        self.scale.scaleFactor.setValue(14)
        self.material = coin.SoMaterial()
        self.sep.addChild(self.pos)
        self.sep.addChild(self.material)
        self.sep.addChild(self.scale)
        self.enabled = True
        self.selected = False

    def setSelected(self, select):
        self.selected = select
        self.sphere.radius = 1.5 if select else 1.0
        self.setEnabled(self.enabled)

    def setEnabled(self, enabled):
        self.enabled = enabled
        if enabled:
            self.material.diffuseColor = self.color[0] if not self.selected else self.color[2]
            self.material.transparency = 0.0
        else:
            self.material.diffuseColor = self.color[1] if not self.selected else self.color[2]
            self.material.transparency = 0.6


class PathDressupTagViewProvider:
    def __init__(self, vobj):
        Path.Log.track()
        self.vobj = vobj
        self.panel = None

        self.debugDisplay()

        # initialized later
        self.obj = None
        self.tags = None
        self.switch = None
        self.colors = None

    def debugDisplay(self):
        # if False and addDebugDisplay():
        #    if not hasattr(self.vobj, 'Debug'):
        #        self.vobj.addProperty('App::PropertyLink', 'Debug', 'Debug', QT_TRANSLATE_NOOP('CAM_DressupTag', 'Some elements for debugging'))
        #        dbg = self.vobj.Object.Document.addObject('App::DocumentObjectGroup', 'TagDebug')
        #        self.vobj.Debug = dbg
        #    return True
        return False

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def setupColors(self):
        def colorForColorValue(val):
            v = [((val >> n) & 0xFF) / 255.0 for n in [24, 16, 8, 0]]
            return coin.SbColor(v[0], v[1], v[2])

        pref = Path.Preferences.preferences()
        #                                                      R         G          B          A
        npc = pref.GetUnsigned("DefaultPathMarkerColor", ((85 * 256 + 255) * 256 + 0) * 256 + 255)
        hpc = pref.GetUnsigned(
            "DefaultHighlightPathColor", ((255 * 256 + 125) * 256 + 0) * 256 + 255
        )
        dpc = pref.GetUnsigned(
            "DefaultDisabledPathColor", ((205 * 256 + 205) * 256 + 205) * 256 + 154
        )
        self.colors = [
            colorForColorValue(npc),
            colorForColorValue(dpc),
            colorForColorValue(hpc),
        ]

    def attach(self, vobj):
        Path.Log.track()
        self.setupColors()
        self.vobj = vobj
        self.obj = vobj.Object
        self.tags = []
        self.switch = coin.SoSwitch()
        vobj.RootNode.addChild(self.switch)
        self.turnMarkerDisplayOn(False)

        if self.obj and self.obj.Base:
            for i in self.obj.Base.InList:
                if hasattr(i, "Group") and self.obj.Base.Name in [o.Name for o in i.Group]:
                    i.Group = [o for o in i.Group if o.Name != self.obj.Base.Name]
            if self.obj.Base.ViewObject:
                self.obj.Base.ViewObject.Visibility = False
            # if self.debugDisplay() and self.vobj.Debug.ViewObject:
            #    self.vobj.Debug.ViewObject.Visibility = False

    def turnMarkerDisplayOn(self, display):
        sw = coin.SO_SWITCH_ALL if display else coin.SO_SWITCH_NONE
        self.switch.whichChild = sw

    def claimChildren(self):
        Path.Log.track()
        # if self.debugDisplay():
        #    return [self.obj.Base, self.vobj.Debug]
        return [self.obj.Base]

    def onDelete(self, arg1=None, arg2=None):
        """this makes sure that the base operation is added back to the job and visible"""
        Path.Log.track()
        if self.obj.Base and self.obj.Base.ViewObject:
            self.obj.Base.ViewObject.Visibility = True
        job = PathUtils.findParentJob(self.obj)
        if arg1.Object and arg1.Object.Base and job:
            job.Proxy.addOperation(arg1.Object.Base, arg1.Object)
            arg1.Object.Base = None
        # if self.debugDisplay():
        #    self.vobj.Debug.removeObjectsFromDocument()
        #    self.vobj.Debug.Document.removeObject(self.vobj.Debug.Name)
        #    self.vobj.Debug = None
        return True

    def updatePositions(self, positions, disabled):
        for tag in self.tags:
            self.switch.removeChild(tag.sep)
        tags = []
        for i, p in enumerate(positions):
            tag = HoldingTagMarker(self.obj.Proxy.pointAtBottom(self.obj, p), self.colors)
            tag.setEnabled(i not in disabled)
            tags.append(tag)
            self.switch.addChild(tag.sep)
        self.tags = tags

    def updateData(self, obj, propName):
        Path.Log.track(propName)
        if "Disabled" == propName:
            self.updatePositions(obj.Positions, obj.Disabled)

    def onModelChanged(self):
        Path.Log.track()
        # if self.debugDisplay():
        #    self.vobj.Debug.removeObjectsFromDocument()
        #    for solid in self.obj.Proxy.solids:
        #        tag = self.obj.Document.addObject('Part::Feature', 'tag')
        #        tag.Shape = solid
        #        if tag.ViewObject and self.vobj.Debug.ViewObject:
        #            tag.ViewObject.Visibility = self.vobj.Debug.ViewObject.Visibility
        #            tag.ViewObject.Transparency = 80
        #        self.vobj.Debug.addObject(tag)
        #    tag.purgeTouched()

    def setEdit(self, vobj, mode=0):
        panel = PathDressupTagTaskPanel(vobj.Object, self)
        self.setupTaskPanel(panel)
        return True

    def unsetEdit(self, vobj, mode):
        if hasattr(self, "panel") and self.panel:
            self.panel.abort()

    def setupTaskPanel(self, panel):
        self.panel = panel
        FreeCADGui.Control.closeDialog()
        FreeCADGui.Control.showDialog(panel)
        panel.setupUi()
        FreeCADGui.Selection.addSelectionGate(self)
        FreeCADGui.Selection.addObserver(self)

    def clearTaskPanel(self):
        self.panel = None
        FreeCADGui.Selection.removeSelectionGate()
        FreeCADGui.Selection.removeObserver(self)
        self.turnMarkerDisplayOn(False)

    # SelectionObserver interface

    def selectTag(self, index):
        Path.Log.track(index)
        for i, tag in enumerate(self.tags):
            tag.setSelected(i == index)

    def tagAtPoint(self, point, matchZ):
        x = point[0]
        y = point[1]
        z = point[2]
        if self.tags and not matchZ:
            z = self.tags[0].point.z
        p = FreeCAD.Vector(x, y, z)
        for i, tag in enumerate(self.tags):
            if Path.Geom.pointsCoincide(p, tag.point, tag.sphere.radius.getValue() * 1.3):
                return i
        return -1

    def allow(self, doc, obj, sub):
        if obj == self.obj:
            return True
        return False

    def addSelection(self, doc, obj, sub, point):
        Path.Log.track(doc, obj, sub, point)
        if hasattr(self, "panel") and self.panel:
            i = self.tagAtPoint(point, sub is None)
            self.panel.selectTagWithId(i)
        FreeCADGui.updateGui()

    def getIcon(self):
        if getattr(PathDressup.baseOp(self.obj), "Active", True):
            return ":/icons/CAM_Dressup.svg"
        else:
            return ":/icons/CAM_OpActive.svg"


def Create(baseObject, name="DressupTag"):
    """
    Create(basePath, name = 'DressupTag') ... create tag dressup object for the given base path.
    Use this command only iff the UI is up - for batch processing see PathDressupTag.Create
    """
    FreeCAD.ActiveDocument.openTransaction("Create a Tag dressup")
    obj = PathDressupTag.Create(baseObject, name)
    obj.ViewObject.Proxy = PathDressupTagViewProvider(obj.ViewObject)
    FreeCAD.ActiveDocument.commitTransaction()
    obj.ViewObject.Document.setEdit(obj.ViewObject, 0)
    return obj


class CommandPathDressupTag:
    def GetResources(self):
        return {
            "Pixmap": "CAM_Dressup",
            "MenuText": QT_TRANSLATE_NOOP("CAM_DressupTag", "Tag"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "CAM_DressupTag", "Creates a tag dress-up object from a selected toolpath"
            ),
        }

    def IsActive(self):
        if FreeCAD.ActiveDocument is not None:
            for o in FreeCAD.ActiveDocument.Objects:
                if o.Name[:3] == "Job":
                    return True
        return False

    def Activated(self):
        # check that the selection contains exactly what we want
        selection = FreeCADGui.Selection.getSelection()
        if len(selection) != 1:
            Path.Log.error(translate("CAM_DressupTag", "Please select one toolpath object") + "\n")
            return
        baseObject = selection[0]

        # everything ok!
        FreeCAD.ActiveDocument.openTransaction("Create Tag Dress-up")
        FreeCADGui.addModule("Path.Dressup.Gui.Tags")
        FreeCADGui.doCommand(
            "Path.Dressup.Gui.Tags.Create(App.ActiveDocument.%s)" % baseObject.Name
        )
        # FreeCAD.ActiveDocument.commitTransaction()  # Final `commitTransaction()` called via TaskPanel.accept()
        FreeCAD.ActiveDocument.recompute()


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand("CAM_DressupTag", CommandPathDressupTag())

Path.Log.notice("Loading PathDressupTagGui... done\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# *****************************************************************************
# *                                                                           *
# *   Copyright (c) 2019 furti <daniel.furtlehner@gmx.net>                    *
# *                                                                           *
# *   This file is part of FreeCAD.                                           *
# *                                                                           *
# *   FreeCAD is free software: you can redistribute it and/or modify it      *
# *   under the terms of the GNU Lesser General Public License as             *
# *   published by the Free Software Foundation, either version 2.1 of the    *
# *   License, or (at your option) any later version.                         *
# *                                                                           *
# *   FreeCAD is distributed in the hope that it will be useful, but          *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of              *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        *
# *   Lesser General Public License for more details.                         *
# *                                                                           *
# *   You should have received a copy of the GNU Lesser General Public        *
# *   License along with FreeCAD. If not, see                                 *
# *   <https://www.gnu.org/licenses/>.                                        *
# *                                                                           *
# *****************************************************************************

# Fence functionality for the Arch Workbench

import math

import FreeCAD
import ArchComponent
import draftobjects.patharray as patharray

if FreeCAD.GuiUp:
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import PySide.QtGui as QtGui
    import FreeCADGui
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond

EAST = FreeCAD.Vector(1, 0, 0)


class _Fence(ArchComponent.Component):
    def __init__(self, obj):

        ArchComponent.Component.__init__(self, obj)
        self.Type = "Fence"
        self.setProperties(obj)
        # Does a IfcType exist?
        # obj.IfcType = "Fence"
        obj.MoveWithHost = False

    def setProperties(self, obj):
        ArchComponent.Component.setProperties(self, obj)

        pl = obj.PropertiesList

        if not "Section" in pl:
            obj.addProperty(
                "App::PropertyLink",
                "Section",
                "Fence",
                QT_TRANSLATE_NOOP("App::Property", "A single section of the fence"),
                locked=True,
            )

        if not "Post" in pl:
            obj.addProperty(
                "App::PropertyLink",
                "Post",
                "Fence",
                QT_TRANSLATE_NOOP("App::Property", "A single fence post"),
                locked=True,
            )

        if not "Path" in pl:
            obj.addProperty(
                "App::PropertyLink",
                "Path",
                "Fence",
                QT_TRANSLATE_NOOP("App::Property", "The Path the fence should follow"),
                locked=True,
            )

        if not "NumberOfSections" in pl:
            obj.addProperty(
                "App::PropertyInteger",
                "NumberOfSections",
                "Fence",
                QT_TRANSLATE_NOOP("App::Property", "The number of sections the fence is built of"),
                locked=True,
            )
            obj.setEditorMode("NumberOfSections", 1)

        if not "NumberOfPosts" in pl:
            obj.addProperty(
                "App::PropertyInteger",
                "NumberOfPosts",
                "Fence",
                QT_TRANSLATE_NOOP("App::Property", "The number of posts used to build the fence"),
                locked=True,
            )
            obj.setEditorMode("NumberOfPosts", 1)

    def dumps(self):
        if hasattr(self, "sectionFaceNumbers"):
            return self.sectionFaceNumbers
        return None

    def loads(self, state):
        if state is not None and isinstance(state, tuple):
            self.sectionFaceNumbers = state[0]
        self.Type = "Fence"

    def execute(self, obj):
        import Part

        pathwire = self.calculatePathWire(obj)

        if not pathwire:
            FreeCAD.Console.PrintLog("ArchFence.execute: path " + obj.Path.Name + " has no edges\n")

            return

        if not obj.Section:
            FreeCAD.Console.PrintLog("ArchFence.execute: Section not set\n")

            return

        if not obj.Post:
            FreeCAD.Console.PrintLog("ArchFence.execute: Post not set\n")

            return

        pathLength = pathwire.Length
        sectionLength = obj.Section.Shape.BoundBox.XMax
        postLength = obj.Post.Shape.BoundBox.XMax

        obj.NumberOfSections = self.calculateNumberOfSections(pathLength, sectionLength, postLength)
        obj.NumberOfPosts = obj.NumberOfSections + 1

        # We assume that the section was drawn in front view.
        # We have to rotate the shape down so that it is aligned
        # correctly by the algorithm later on
        downRotation = FreeCAD.Rotation(FreeCAD.Vector(1, 0, 0), -90)

        postPlacements = self.calculatePostPlacements(obj, pathwire, downRotation)

        postShapes = self.calculatePosts(obj, postPlacements)
        sectionShapes, sectionFaceNumbers = self.calculateSections(
            obj, postPlacements, postLength, sectionLength
        )

        allShapes = []
        allShapes.extend(postShapes)
        allShapes.extend(sectionShapes)

        compound = Part.makeCompound(allShapes)

        self.sectionFaceNumbers = sectionFaceNumbers

        obj.Shape = compound

    def calculateNumberOfSections(self, pathLength, sectionLength, postLength):
        realSectionLength = sectionLength + postLength

        return math.ceil(pathLength / realSectionLength)

    def calculatePostPlacements(self, obj, pathwire, rotation):
        postWidth = obj.Post.Shape.BoundBox.YMax

        # We want to center the posts on the path. So move them the half width in
        transformationVector = FreeCAD.Vector(0, -postWidth / 2, 0)

        return patharray.placements_on_path(
            rotation, pathwire, obj.NumberOfPosts, transformationVector, True
        )

    def calculatePosts(self, obj, postPlacements):
        posts = []

        for placement in postPlacements:
            postCopy = obj.Post.Shape.copy()
            postCopy.Placement = placement

            posts.append(postCopy)

        return posts

    def calculateSections(self, obj, postPlacements, postLength, sectionLength):
        import Part

        shapes = []

        # For the colorization algorithm we have to store the number of faces for each section
        # It is possible that a section is clipped. Then the number of faces is not equal to the
        # number of faces in the original section
        faceNumbers = []

        for i in range(obj.NumberOfSections):
            startPlacement = postPlacements[i]
            endPlacement = postPlacements[i + 1]

            sectionLine = Part.LineSegment(startPlacement.Base, endPlacement.Base)
            sectionBase = sectionLine.value(postLength)

            if startPlacement.Rotation.isSame(endPlacement.Rotation):
                sectionRotation = endPlacement.Rotation
            else:
                direction = endPlacement.Base.sub(startPlacement.Base)

                sectionRotation = FreeCAD.Rotation(EAST, direction)

            placement = FreeCAD.Placement()
            placement.Base = sectionBase
            placement.Rotation = sectionRotation

            sectionCopy = obj.Section.Shape.copy()

            if sectionLength > sectionLine.length() - postLength:
                # Part.show(Part.Shape([sectionLine]), 'line')
                sectionCopy = self.clipSection(
                    sectionCopy, sectionLength, sectionLine.length() - postLength
                )

            sectionCopy = Part.Compound(
                [sectionCopy]
            )  # nest in compound to ensure correct Placement
            sectionCopy.Placement = placement

            shapes.append(sectionCopy)
            faceNumbers.append(len(sectionCopy.Faces))

        return (shapes, faceNumbers)

    def clipSection(self, shape, length, clipLength):
        import Part

        boundBox = shape.BoundBox
        lengthToCut = length - clipLength
        halfLengthToCut = lengthToCut / 2

        leftBox = Part.makeBox(
            halfLengthToCut,
            boundBox.YMax + 1,
            boundBox.ZMax + 1,
            FreeCAD.Vector(boundBox.XMin, boundBox.YMin, boundBox.ZMin),
        )
        rightBox = Part.makeBox(
            halfLengthToCut,
            boundBox.YMax + 1,
            boundBox.ZMax + 1,
            FreeCAD.Vector(
                boundBox.XMin + halfLengthToCut + clipLength, boundBox.YMin, boundBox.ZMin
            ),
        )

        newShape = shape.cut([leftBox, rightBox])
        newBoundBox = newShape.BoundBox

        newShape.translate(FreeCAD.Vector(-newBoundBox.XMin, 0, 0))

        return newShape.removeSplitter()

    def calculatePathWire(self, obj):
        if hasattr(obj.Path.Shape, "Wires") and obj.Path.Shape.Wires:
            return obj.Path.Shape.Wires[0]
        elif obj.Path.Shape.Edges:
            return Part.Wire(obj.Path.Shape.Edges)

        return None


class _ViewProviderFence(ArchComponent.ViewProviderComponent):
    "A View Provider for the Fence object"

    def __init__(self, vobj):
        ArchComponent.ViewProviderComponent.__init__(self, vobj)
        # setProperties of ArchComponent will be overwritten
        # thus setProperties from ArchComponent will be explicit called to get the properties
        ArchComponent.ViewProviderComponent.setProperties(self, vobj)
        self.setProperties(vobj)

    def setProperties(self, vobj):
        pl = vobj.PropertiesList

        if not "UseOriginalColors" in pl:
            vobj.addProperty(
                "App::PropertyBool",
                "UseOriginalColors",
                "Fence",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "When true, the fence will be colored like the original post and section.",
                ),
                locked=True,
            )

    def attach(self, vobj):
        self.setProperties(vobj)

        return super().attach(vobj)

    def getIcon(self):
        import Arch_rc

        return ":/icons/Arch_Fence_Tree.svg"

    def claimChildren(self):
        children = []

        if self.Object.Section:
            children.append(self.Object.Section)

        if self.Object.Post:
            children.append(self.Object.Post)

        if self.Object.Path:
            children.append(self.Object.Path)

        return children

    def updateData(self, obj, prop):
        colorProps = ["Shape", "Section", "Post", "Path"]

        if prop in colorProps:
            self.applyColors(obj)
        else:
            super().updateData(obj, prop)

    def onChanged(self, vobj, prop):
        if prop == "UseOriginalColors":
            self.applyColors(vobj.Object)
        else:
            super().onChanged(vobj, prop)

    def applyColors(self, obj):
        # Note that the clipSection function changes the face numbering of the
        # fence section. This happens even if the total number of faces does not
        # change. If UseOriginalColors is True, the end result of this function
        # will only be correct if all faces of the section have the same color.

        vobj = obj.ViewObject
        if not vobj.UseOriginalColors:
            vobj.ShapeAppearance = [vobj.ShapeAppearance[0]]
        else:
            post = obj.Post
            section = obj.Section

            # If post and/or section are Std_Parts they may not have a Shape attr (yet):
            if not hasattr(post, "Shape"):
                return
            if not hasattr(section, "Shape"):
                return

            numberOfPostFaces = len(post.Shape.Faces)
            numberOfSectionFaces = len(section.Shape.Faces)

            if hasattr(obj.Proxy, "sectionFaceNumbers"):
                sectionFaceNumbers = obj.Proxy.sectionFaceNumbers
            else:
                sectionFaceNumbers = [0]

            if numberOfPostFaces == 0 or sum(sectionFaceNumbers) == 0:
                return

            postColors = self.normalizeColors(post, numberOfPostFaces)
            defaultSectionColors = self.normalizeColors(section, numberOfSectionFaces)

            ownColors = []

            # At first all posts are added to the shape
            for i in range(obj.NumberOfPosts):
                ownColors.extend(postColors)

            # Next all sections are added
            for i in range(obj.NumberOfSections):
                actualSectionFaceCount = sectionFaceNumbers[i]

                if actualSectionFaceCount == numberOfSectionFaces:
                    ownColors.extend(defaultSectionColors)
                else:
                    ownColors.extend(self.normalizeColors(section, actualSectionFaceCount))

            vobj.DiffuseColor = ownColors

    def normalizeColors(self, obj, numberOfFaces):
        if obj.TypeId == "PartDesign::Body":
            # When colorizing a PartDesign Body we have two options
            # 1. The whole body got a shape color, that means the tip has only a single diffuse color set
            #   so we use the shape color of the body
            # 2. "Set colors" was called on the tip and the individual faces where colorized.
            #   We use the diffuseColors of the tip in that case
            if len(obj.Tip.ViewObject.DiffuseColor) > 1:
                colors = obj.Tip.ViewObject.DiffuseColor
            else:
                colors = obj.ViewObject.DiffuseColor
        else:
            import Draft

            colors = Draft.get_diffuse_color(obj)  # To handle Std_Parts for example.

        numberOfColors = len(colors)

        if numberOfColors == 1:
            return colors * numberOfFaces

        if numberOfColors == numberOfFaces:
            return colors

        # It is possible, that we have fewer faces than colors when something
        # got clipped. Remove the unneeded colors at the beginning and end.

        # Even if clipSection did not change the face numbering this code would
        # not work properly.
        halfNumberOfFacesToRemove = (numberOfColors - numberOfFaces) / 2
        start = int(math.ceil(halfNumberOfFacesToRemove))
        end = start + numberOfFaces
        return colors[start:end]


def hide(obj):
    if hasattr(obj, "ViewObject") and obj.ViewObject:
        obj.ViewObject.Visibility = False



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides various functions to work with wires."""
## @package wires
# \ingroup draftgeoutils
# \brief Provides various functions to work with wires.

import math
import lazy_loader.lazy_loader as lz

import FreeCAD as App
import DraftVecUtils
import WorkingPlane

from draftgeoutils.general import geomType, vec, precision
from draftgeoutils.geometry import get_normal
from draftgeoutils.geometry import project_point_on_plane
from draftgeoutils.edges import findMidpoint, isLine

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")

## \addtogroup draftgeoutils
# @{


def findWires(edgeslist):
    """Find wires in a list of edges."""
    return [Part.Wire(e) for e in Part.sortEdges(edgeslist)]


def findWiresOld2(edgeslist):
    """Find connected wires in the given list of edges."""

    def touches(e1, e2):
        """Return True if two edges connect at the edges."""
        if len(e1.Vertexes) < 2:
            return False
        if len(e2.Vertexes) < 2:
            return False
        if DraftVecUtils.equals(e1.Vertexes[0].Point, e2.Vertexes[0].Point):
            return True
        if DraftVecUtils.equals(e1.Vertexes[0].Point, e2.Vertexes[-1].Point):
            return True
        if DraftVecUtils.equals(e1.Vertexes[-1].Point, e2.Vertexes[0].Point):
            return True
        if DraftVecUtils.equals(e1.Vertexes[-1].Point, e2.Vertexes[-1].Point):
            return True
        return False

    edges = edgeslist[:]
    wires = []
    lost = []
    while edges:
        e = edges[0]
        if not wires:
            # create first group
            edges.remove(e)
            wires.append([e])
        else:
            found = False
            for w in wires:
                if not found:
                    for we in w:
                        if touches(e, we):
                            edges.remove(e)
                            w.append(e)
                            found = True
                            break
            if not found:
                if e in lost:
                    # we already tried this edge, and still nothing
                    edges.remove(e)
                    wires.append([e])
                    lost = []
                else:
                    # put to the end of the list
                    edges.remove(e)
                    edges.append(e)
                    lost.append(e)
    nwires = []
    for w in wires:
        try:
            wi = Part.Wire(w)
        except Part.OCCError:
            print("couldn't join some edges")
        else:
            nwires.append(wi)
    return nwires


def findWiresOld(edges):
    """Return a list of lists containing edges that can be connected.

    Find connected edges in the list.
    """
    raise DeprecationWarning(
        "This function shouldn't be called anymore. " "Use findWires() instead"
    )

    def verts(shape):
        return [shape.Vertexes[0].Point, shape.Vertexes[-1].Point]

    def group(shapes):
        shapesIn = shapes[:]
        shapesOut = [shapesIn.pop()]
        changed = False
        for s in shapesIn:
            if len(s.Vertexes) < 2:
                continue
            else:
                clean = True
                for v in verts(s):
                    for i in range(len(shapesOut)):
                        if clean and (v in verts(shapesOut[i])):
                            shapesOut[i] = Part.Wire(shapesOut[i].Edges + s.Edges)
                            changed = True
                            clean = False
                if clean:
                    shapesOut.append(s)
        return changed, shapesOut

    working = True
    edgeSet = edges

    while working:
        result = group(edgeSet)
        working = result[0]
        edgeSet = result[1]

    return result[1]


def flattenWire(wire, origin=None, normal=None):
    """Force a wire to be flat on a plane defined by an origin and a normal.

    If origin or normal are None they are derived from the wire.
    """
    if normal is None:
        normal = get_normal(wire)
        # for backward compatibility with previous getNormal implementation
        if normal is None:
            normal = App.Vector(0, 0, 1)
    if origin is None:
        origin = wire.Vertexes[0].Point

    points = [project_point_on_plane(vert.Point, origin, normal) for vert in wire.Vertexes]
    if wire.isClosed():
        points.append(points[0])
    new_wire = Part.makePolygon(points)

    return new_wire


def superWire(edgeslist, closed=False):
    """Force a wire between edges that don't have coincident endpoints.

    Forces a wire between edges that don't necessarily
    have coincident endpoints. If closed=True, the wire will always be closed.
    """

    def median(v1, v2):
        vd = v2.sub(v1)
        vd.scale(0.5, 0.5, 0.5)
        return v1.add(vd)

    edges = Part.__sortEdges__(edgeslist)
    print(edges)
    newedges = []

    for i in range(len(edges)):
        curr = edges[i]
        if i == 0:
            if closed:
                prev = edges[-1]
            else:
                prev = None
        else:
            prev = edges[i - 1]

        if i == (len(edges) - 1):
            if closed:
                _next = edges[0]
            else:
                _next = None
        else:
            _next = edges[i + 1]

        print(i, prev, curr, _next)

        if prev:
            if curr.Vertexes[0].Point == prev.Vertexes[-1].Point:
                p1 = curr.Vertexes[0].Point
            else:
                p1 = median(curr.Vertexes[0].Point, prev.Vertexes[-1].Point)
        else:
            p1 = curr.Vertexes[0].Point

        if _next:
            if curr.Vertexes[-1].Point == _next.Vertexes[0].Point:
                p2 = _next.Vertexes[0].Point
            else:
                p2 = median(curr.Vertexes[-1].Point, _next.Vertexes[0].Point)
        else:
            p2 = curr.Vertexes[-1].Point

        if geomType(curr) == "Line":
            print("line", p1, p2)
            newedges.append(Part.LineSegment(p1, p2).toShape())
        elif geomType(curr) == "Circle":
            p3 = findMidpoint(curr)
            print("arc", p1, p3, p2)
            newedges.append(Part.Arc(p1, p3, p2).toShape())
        else:
            print("Cannot superWire edges that are not lines or arcs")
            return None

    print(newedges)
    return Part.Wire(newedges)


def isReallyClosed(wire):
    if isinstance(wire, (Part.Wire, Part.Edge)):
        return wire.isClosed()
    return isinstance(wire, Part.Face)


def curvetowire(obj, steps):
    """Discretize the object and return a list of edges."""
    points = obj.copy().discretize(steps)
    p0 = points[0]
    edgelist = []
    for p in points[1:]:
        edge = Part.makeLine((p0.x, p0.y, p0.z), (p.x, p.y, p.z))
        edgelist.append(edge)
        p0 = p
    return edgelist


def curvetosegment(curve, seglen):
    """Discretize the curve and return a list of edges."""
    points = curve.discretize(seglen)
    p0 = points[0]
    edgelist = []
    for p in points[1:]:
        edge = Part.makeLine((p0.x, p0.y, p0.z), (p.x, p.y, p.z))
        edgelist.append(edge)
        p0 = p
    return edgelist


def rebaseWire(wire, vidx=0):
    """Return a copy of the wire with the first vertex indicated by the index.

    Return a new wire which is a copy of the current wire,
    but where the first vertex is the vertex indicated by the given
    index vidx, starting from 1.
    0 will return an exact copy of the wire.
    """
    if vidx < 1:
        return wire

    if vidx > len(wire.Vertexes):
        # print("Vertex index above maximum")
        return wire

    # This can be done in one step
    return Part.Wire(wire.Edges[vidx - 1 :] + wire.Edges[: vidx - 1])


def removeInterVertices(wire):
    """Remove middle vertices from a straight wire and return a new wire.

    Remove unneeded vertices, those that are in the middle of a straight line,
    from a wire, return a new wire.
    """
    _pre = precision()
    edges = Part.__sortEdges__(wire.Edges)
    nverts = []

    def getvec(v1, v2):
        if not abs(round(v1.getAngle(v2), _pre) in [0, round(math.pi, _pre)]):
            nverts.append(edges[i].Vertexes[-1].Point)

    for i in range(len(edges) - 1):
        vA = vec(edges[i])
        vB = vec(edges[i + 1])
        getvec(vA, vB)

    vA = vec(edges[-1])
    vB = vec(edges[0])
    getvec(vA, vB)

    if nverts:
        if wire.isClosed():
            nverts.append(nverts[0])
        w = Part.makePolygon(nverts)
        return w
    else:
        return wire


def cleanProjection(shape, tessellate=True, seglength=0.05):
    """Return a compound of edges, optionally tessellate ellipses, splines
    and bezcurves.

    The function was formerly used to workaround bugs in the projection
    algorithm. These bugs have since been fixed. Now the function is only
    used when tessellation of ellipses, splines and bezcurves is required
    (DXF output and Draft_Shape2DView).
    """
    oldedges = shape.Edges
    newedges = []
    for e in oldedges:
        typ = geomType(e)
        try:
            if typ in ["Line", "Circle"]:
                newedges.append(e)
            elif typ == "Ellipse":
                if tessellate:
                    newedges.append(Part.Wire(curvetowire(e, seglength)))
                else:
                    newedges.append(e)
            elif typ in ["BSplineCurve", "BezierCurve"]:
                if isLine(e.Curve):
                    line = Part.LineSegment(e.Vertexes[0].Point, e.Vertexes[-1].Point)
                    newedges.append(line)
                elif tessellate:
                    newedges.append(Part.Wire(curvetowire(e, seglength)))
                else:
                    newedges.append(e)
            else:
                newedges.append(e)
        except Part.OCCError:
            print("Debug: error cleaning edge ", e)

    return Part.makeCompound(newedges)


def tessellateProjection(shape, seglen):
    """Return projection with BSplines and Ellipses broken into line segments.

    Useful for exporting projected views to DXF files.
    """
    oldedges = shape.Edges
    newedges = []
    for e in oldedges:
        try:
            if geomType(e) == "Line":
                newedges.append(e.Curve.toShape())
            elif geomType(e) == "Circle":
                newedges.append(e.Curve.toShape())
            elif geomType(e) == "Ellipse":
                newedges.append(Part.Wire(curvetosegment(e, seglen)))
            elif geomType(e) == "BSplineCurve":
                newedges.append(Part.Wire(curvetosegment(e, seglen)))
            else:
                newedges.append(e)
        except Part.OCCError:
            print("Debug: error cleaning edge ", e)

    return Part.makeCompound(newedges)


def get_placement_perpendicular_to_wire(wire):
    """Return the placement whose base is the wire's first vertex and it's z axis aligned to the wire's tangent."""
    pl = App.Placement()
    if wire.Length > 0.0:
        pl.Base = wire.OrderedVertexes[0].Point
        first_edge = wire.OrderedEdges[0]
        if first_edge.Orientation == "Forward":
            zaxis = -first_edge.tangentAt(first_edge.FirstParameter)
        else:
            zaxis = first_edge.tangentAt(first_edge.LastParameter)
        pl.Rotation = App.Rotation(App.Vector(1, 0, 0), App.Vector(0, 0, 1), zaxis, "ZYX")
    else:
        App.Console.PrintError(
            "debug: get_placement_perpendicular_to_wire called with a zero-length wire.\n"
        )
    return pl


def get_extended_wire(wire, offset_start, offset_end):
    """Return a wire trimmed (negative offset) or extended (positive offset) at its first vertex, last vertex or both ends.

    get_extended_wire(wire, -100.0, 0.0) -> returns a copy of the wire with its first 100 mm removed
    get_extended_wire(wire, 0.0, 100.0) -> returns a copy of the wire extended by 100 mm after it's last vertex
    """
    if min(offset_start, offset_end, offset_start + offset_end) <= -wire.Length:
        App.Console.PrintError(
            "debug: get_extended_wire error, wire's length insufficient for trimming.\n"
        )
        return wire
    if offset_start < 0:  # Trim the wire from the first vertex
        offset_start = -offset_start
        out_edges = []
        for edge in wire.OrderedEdges:
            if offset_start >= edge.Length:  # Remove entire edge
                offset_start -= edge.Length
            elif round(offset_start, precision()) > 0:  # Split edge, to remove the required length
                if edge.Orientation == "Forward":
                    new_edge = edge.split(edge.getParameterByLength(offset_start)).OrderedEdges[1]
                else:
                    new_edge = edge.split(
                        edge.getParameterByLength(edge.Length - offset_start)
                    ).OrderedEdges[0]
                new_edge.Placement = (
                    edge.Placement
                )  # Strangely, edge.split discards the placement and orientation
                new_edge.Orientation = edge.Orientation
                out_edges.append(new_edge)
                offset_start = 0
            else:  # Keep the remaining entire edges
                out_edges.append(edge)
        wire = Part.Wire(out_edges)
    elif offset_start > 0:  # Extend the first edge along its normal
        first_edge = wire.OrderedEdges[0]
        if first_edge.Orientation == "Forward":
            start, end = first_edge.FirstParameter, first_edge.LastParameter
            vec = first_edge.tangentAt(start).multiply(offset_start)
        else:
            start, end = first_edge.LastParameter, first_edge.FirstParameter
            vec = -first_edge.tangentAt(start).multiply(offset_start)
        if geomType(first_edge) == "Line":  # Replace first edge with the extended new edge
            new_edge = Part.LineSegment(
                first_edge.valueAt(start).sub(vec), first_edge.valueAt(end)
            ).toShape()
            wire = Part.Wire([new_edge] + wire.OrderedEdges[1:])
        else:  # Add a straight edge before the first vertex
            new_edge = Part.LineSegment(
                first_edge.valueAt(start).sub(vec), first_edge.valueAt(start)
            ).toShape()
            wire = Part.Wire([new_edge] + wire.OrderedEdges)
    if offset_end < 0:  # Trim the wire from the last vertex
        offset_end = -offset_end
        out_edges = []
        for edge in reversed(wire.OrderedEdges):
            if offset_end >= edge.Length:  # Remove entire edge
                offset_end -= edge.Length
            elif round(offset_end, precision()) > 0:  # Split edge, to remove the required length
                if edge.Orientation == "Forward":
                    new_edge = edge.split(
                        edge.getParameterByLength(edge.Length - offset_end)
                    ).OrderedEdges[0]
                else:
                    new_edge = edge.split(edge.getParameterByLength(offset_end)).OrderedEdges[1]
                new_edge.Placement = (
                    edge.Placement
                )  # Strangely, edge.split discards the placement and orientation
                new_edge.Orientation = edge.Orientation
                out_edges.insert(0, new_edge)
                offset_end = 0
            else:  # Keep the remaining entire edges
                out_edges.insert(0, edge)
        wire = Part.Wire(out_edges)
    elif offset_end > 0:  # Extend the last edge along its normal
        last_edge = wire.OrderedEdges[-1]
        if last_edge.Orientation == "Forward":
            start, end = last_edge.FirstParameter, last_edge.LastParameter
            vec = last_edge.tangentAt(end).multiply(offset_end)
        else:
            start, end = last_edge.LastParameter, last_edge.FirstParameter
            vec = -last_edge.tangentAt(end).multiply(offset_end)
        if geomType(last_edge) == "Line":  # Replace last edge with the extended new edge
            new_edge = Part.LineSegment(
                last_edge.valueAt(start), last_edge.valueAt(end).add(vec)
            ).toShape()
            wire = Part.Wire(wire.OrderedEdges[:-1] + [new_edge])
        else:  # Add a straight edge after the last vertex
            new_edge = Part.LineSegment(
                last_edge.valueAt(end), last_edge.valueAt(end).add(vec)
            ).toShape()
            wire = Part.Wire(wire.OrderedEdges + [new_edge])
    return wire


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import re
import pathlib
from typing import List, Dict, Tuple, Optional, cast
from ..uri import AssetUri
from .base import AssetStore


def _resolve_case_insensitive(path: pathlib.Path) -> pathlib.Path:
    if path.is_file():
        return path

    # pathlib in Python 2.10 (2.1x versions) does not support
    # 'case_sensitive' argument in glob. Instead, simulate
    # case-insensitive search manually
    pattern = path.name.lower()
    for p in path.parent.glob("*"):
        if p.name.lower() == pattern:
            return p
    return path

    # Use this starting Python 2.13:
    try:
        return next(path.parent.glob(path.name, case_sensitive=False))
    except StopIteration:
        return path


class FileStore(AssetStore):
    """
    Asset store implementation for the local filesystem with optional
    versioning.

    Maps URIs of the form <asset_type>://<asset_id>[/<version>]
    to paths within a base directory.

    The mapping to file system paths is configurable depending on the asset
    type. Example mapping:

        mapping = {
            "*": "{asset_type}/{asset_id}/{version}.dat",
            "model": "models_dir/{asset_id}-{version}.ml",
            "dataset": "data/{asset_id}.csv" # Unversioned (conceptual version "1")
        }

    Placeholders like {version} are matched greedily (.*), but for compatibility,
    versions are expected to be numeric strings for versioned assets.
    """

    DEFAULT_MAPPING = {
        # Default from original problem doc was "{asset_type}/{asset_id}/{id}/<version>"
        # Adjusted to a more common simple case:
        "*": "{asset_type}/{asset_id}/{version}"
    }

    KNOWN_PLACEHOLDERS = {"asset_type", "asset_id", "id", "version"}

    def __init__(
        self,
        name: str,
        base_dir: pathlib.Path,
        mapping: Optional[Dict[str, str]] = None,
    ):
        super().__init__(name)
        self._base_dir = base_dir.resolve()
        self._mapping = mapping if mapping is not None else self.DEFAULT_MAPPING.copy()
        self._validate_patterns_on_init()
        # For _path_to_uri: iterate specific keys before '*' to ensure correct pattern matching
        self._sorted_mapping_keys = sorted(self._mapping.keys(), key=lambda k: (k == "*", k))

    def _validate_patterns_on_init(self):
        if not self._mapping:
            raise ValueError("Asset store mapping cannot be empty.")

        for asset_type_key, path_format in self._mapping.items():
            if not isinstance(path_format, str):
                raise TypeError(f"Path format for key '{asset_type_key}' must be a string.")

            placeholders_in_format = set(re.findall(r"\{([^}]+)\}", path_format))
            for ph_name in placeholders_in_format:
                if ph_name not in self.KNOWN_PLACEHOLDERS:
                    raise ValueError(
                        f"Unknown placeholder {{{ph_name}}} in pattern: '{path_format}'. Allowed: {self.KNOWN_PLACEHOLDERS}"
                    )

            has_asset_id_ph = "asset_id" in placeholders_in_format or "id" in placeholders_in_format
            if not has_asset_id_ph:
                raise ValueError(
                    f"Pattern '{path_format}' for key '{asset_type_key}' must include {{asset_id}} or {{id}}."
                )

            # CORRECTED LINE: Check for the placeholder name "asset_type" not "{asset_type}"
            if asset_type_key == "*" and "asset_type" not in placeholders_in_format:
                raise ValueError(
                    f"Pattern '{path_format}' for wildcard key '*' must include {{asset_type}}."
                )

    @staticmethod
    def _match_path_to_format_string(format_str: str, path_str_posix: str) -> Dict[str, str]:
        """Matches a POSIX-style path string against a format string."""
        tokens = re.split(r"\{(.*?)\}", format_str)  # format_str uses /
        if len(tokens) == 1:  # No placeholders
            if format_str == path_str_posix:
                return {}
            raise ValueError(f"Path '{path_str_posix}' does not match pattern '{format_str}'")

        keywords = tokens[1::2]
        regex_parts = []
        for i, literal_part in enumerate(tokens[0::2]):
            # Literal parts from format_str (using /) are escaped.
            # The path_str_posix is already normalized to /, so direct matching works.
            regex_parts.append(re.escape(literal_part))
            if i < len(keywords):
                regex_parts.append(f"(?P<{keywords[i]}>.*)")

        pattern_regex_str = "".join(regex_parts)
        match_obj = re.fullmatch(pattern_regex_str, path_str_posix, re.I)

        if not match_obj:
            raise ValueError(
                f"Path '{path_str_posix}' does not match format '{format_str}' (regex: '{pattern_regex_str}')"
            )
        return {kw: match_obj.group(kw) for kw in keywords}

    def _get_path_format_for_uri_type(self, uri_asset_type: str) -> str:
        if uri_asset_type in self._mapping:
            return self._mapping[uri_asset_type]
        if "*" in self._mapping:
            return self._mapping["*"]
        raise ValueError(
            f"No mapping pattern for asset_type '{uri_asset_type}' and no '*' fallback."
        )

    def _path_to_uri(self, file_path: pathlib.Path) -> Optional[AssetUri]:
        """Converts a filesystem path to an AssetUri, if it matches a pattern."""
        if not file_path.is_file():
            return None

        try:
            # Convert to relative path object first, then to POSIX string for matching
            relative_path_obj = file_path.relative_to(self._base_dir)
            relative_path_posix = relative_path_obj.as_posix()
        except ValueError:
            return None  # Path not under base_dir

        for asset_type_key in self._sorted_mapping_keys:
            path_format_str = self._mapping[asset_type_key]  # Pattern uses /
            try:
                components = FileStore._match_path_to_format_string(
                    path_format_str, relative_path_posix
                )

                asset_id = components.get("asset_id", components.get("id"))
                if not asset_id:
                    continue

                current_asset_type: str
                if "{asset_type}" in path_format_str:
                    current_asset_type = components.get("asset_type", "")
                    if not current_asset_type:
                        continue
                else:
                    current_asset_type = asset_type_key
                    if current_asset_type == "*":
                        continue  # Invalid state, caught by validation

                version_str: str
                if "{version}" in path_format_str:
                    version_str = components.get("version", "")
                    if not version_str or not version_str.isdigit():
                        continue
                else:
                    version_str = "1"

                return AssetUri.build(
                    asset_type=current_asset_type,
                    asset_id=asset_id,
                    version=version_str,
                )
            except ValueError:  # No match
                continue
        return None

    def set_dir(self, new_dir: pathlib.Path):
        """Sets the base directory for the store."""
        self._base_dir = new_dir.resolve()

    def _uri_to_path(self, uri: AssetUri) -> pathlib.Path:
        """Converts an AssetUri to a filesystem path using mapping."""
        path_format_str = self._get_path_format_for_uri_type(uri.asset_type)

        format_values: Dict[str, str] = {
            "asset_type": uri.asset_type,
            "asset_id": uri.asset_id,
            "id": uri.asset_id,
        }

        # Only add 'version' to format_values if the pattern expects it AND uri.version is set.
        # uri.version must be a string for .format() (e.g. "1", not None).
        if "{version}" in path_format_str:
            if uri.version is None:
                # This state implies an issue: a versioned pattern is being used
                # but the URI hasn't had its version appropriately set (e.g. to "1" for create,
                # or resolved from "latest").
                raise ValueError(
                    f"URI version is None for versioned pattern '{path_format_str}'. URI: {uri}"
                )
            format_values["version"] = uri.version

        try:
            # Patterns use '/', pathlib handles OS-specific path construction.
            resolved_path_str = path_format_str.format(**format_values)
        except KeyError as e:
            raise ValueError(
                f"Pattern '{path_format_str}' placeholder {{{e}}} missing in URI data for {uri}."
            )

        return self._base_dir / resolved_path_str

    async def get(self, uri: AssetUri) -> bytes:
        """Retrieve the raw byte data for the asset at the given URI."""
        path_to_read: pathlib.Path

        if uri.version == "latest":
            query_uri = AssetUri.build(
                asset_type=uri.asset_type,
                asset_id=uri.asset_id,
                params=uri.params,
            )
            versions = await self.list_versions(query_uri)
            if not versions:
                raise FileNotFoundError(f"No versions found for {uri.asset_type}://{uri.asset_id}")
            latest_version_uri = versions[-1]  # list_versions now returns AssetUri with params
            path_to_read = self._uri_to_path(latest_version_uri)
        else:
            request_uri = uri
            path_format_str = self._get_path_format_for_uri_type(uri.asset_type)
            is_versioned_pattern = "{version}" in path_format_str

            if not is_versioned_pattern:
                if uri.version is not None and uri.version != "1":
                    raise FileNotFoundError(
                        f"Asset type '{uri.asset_type}' is unversioned. "
                        f"Version '{uri.version}' invalid for URI {uri}. Use '1' or no version."
                    )
                if uri.version is None:  # Conceptual "type://id" -> "type://id/1"
                    request_uri = AssetUri.build(
                        asset_type=uri.asset_type,
                        asset_id=uri.asset_id,
                        version="1",
                        params=uri.params,
                    )
            elif (
                uri.version is None
            ):  # Versioned pattern but URI has version=None (and not "latest")
                raise FileNotFoundError(
                    f"Version required for asset type '{uri.asset_type}' (pattern: '{path_format_str}'). URI: {uri}"
                )
            path_to_read = self._uri_to_path(request_uri)

        path_to_read = _resolve_case_insensitive(path_to_read)
        try:
            with open(path_to_read, mode="rb") as f:
                return f.read()
        except FileNotFoundError:
            raise FileNotFoundError(f"Asset for URI {uri} not found at path {path_to_read}")
        except IsADirectoryError:
            raise FileNotFoundError(f"Asset URI {uri} resolved to a directory: {path_to_read}")

    async def exists(self, uri: AssetUri) -> bool:
        """Check if the asset exists at the given URI."""
        try:
            await self.get(uri)
            return True
        except FileNotFoundError:
            return False
        except IsADirectoryError:
            # If the path is a directory, it means the asset exists but is not a file.
            return False

    async def delete(self, uri: AssetUri) -> None:
        """Delete the asset at the given URI."""
        paths_to_delete: List[pathlib.Path] = []
        parent_dirs_of_deleted_files = set()  # To track for cleanup

        path_format_str = self._get_path_format_for_uri_type(uri.asset_type)
        is_versioned_pattern = "{version}" in path_format_str

        if uri.version is None:  # Delete all versions or the single unversioned file
            for path_obj in self._base_dir.rglob("*"):
                parsed_uri = self._path_to_uri(path_obj)
                if (
                    parsed_uri
                    and parsed_uri.asset_type == uri.asset_type
                    and parsed_uri.asset_id == uri.asset_id
                ):
                    paths_to_delete.append(path_obj)
        else:  # Delete a specific version or an unversioned file (if version is "1")
            target_uri_for_path = uri
            if not is_versioned_pattern:
                if uri.version != "1":
                    return  # Idempotent: non-"1" version of unversioned asset "deleted"
                target_uri_for_path = AssetUri.build(
                    asset_type=uri.asset_type,
                    asset_id=uri.asset_id,
                    version="1",
                    params=uri.params,
                )

            path = self._uri_to_path(target_uri_for_path)
            path = _resolve_case_insensitive(path)
            if path.is_file():
                paths_to_delete.append(path)

        for p_del in paths_to_delete:
            try:
                p_del.unlink()
                parent_dirs_of_deleted_files.add(p_del.parent)
            except FileNotFoundError:
                pass

        # Clean up empty parent directories, from deepest first
        sorted_parents = sorted(
            list(parent_dirs_of_deleted_files),
            key=lambda p: len(p.parts),
            reverse=True,
        )
        for parent_dir in sorted_parents:
            current_cleanup_path = parent_dir
            while (
                current_cleanup_path.exists()
                and current_cleanup_path.is_dir()
                and current_cleanup_path != self._base_dir
                and current_cleanup_path.is_relative_to(self._base_dir)
                and not any(current_cleanup_path.iterdir())
            ):
                try:
                    current_cleanup_path.rmdir()
                    current_cleanup_path = current_cleanup_path.parent
                except OSError:
                    break

    async def create(self, asset_type: str, asset_id: str, data: bytes) -> AssetUri:
        """Create a new asset in the store with the given data."""
        # New assets are conceptually version "1"
        uri_to_create = AssetUri.build(asset_type=asset_type, asset_id=asset_id, version="1")
        asset_path = self._uri_to_path(uri_to_create)

        if asset_path.exists():
            # More specific error messages based on what exists
            if asset_path.is_file():
                raise FileExistsError(f"Asset file already exists at {asset_path}")
            if asset_path.is_dir():
                raise IsADirectoryError(f"A directory exists at target path {asset_path}")
            raise FileExistsError(f"Path {asset_path} already exists (unknown type).")

        asset_path.parent.mkdir(parents=True, exist_ok=True)
        with open(asset_path, mode="wb") as f:
            f.write(data)
        return uri_to_create

    async def update(self, uri: AssetUri, data: bytes) -> AssetUri:
        """Update the asset at the given URI with new data, creating a new version."""
        # Get a Uri without the version number, use it to find all versions.
        query_uri = AssetUri.build(
            asset_type=uri.asset_type, asset_id=uri.asset_id, params=uri.params
        )
        existing_versions = await self.list_versions(query_uri)
        if not existing_versions:
            raise FileNotFoundError(
                f"No versions for asset {uri.asset_type}://{uri.asset_id} to update."
            )

        # Create a Uri for the NEXT version number.
        latest_version_uri = existing_versions[-1]
        latest_version_num = int(cast(str, latest_version_uri.version))
        next_version_str = str(latest_version_num + 1)
        next_uri = AssetUri.build(
            asset_type=uri.asset_type,
            asset_id=uri.asset_id,
            version=next_version_str,
            params=uri.params,
        )
        asset_path = self._uri_to_path(next_uri)
        asset_path = _resolve_case_insensitive(asset_path)

        # If the file is versioned, then the new version should not yet exist.
        # Double check to be sure.
        path_format_str = self._get_path_format_for_uri_type(uri.asset_type)
        is_versioned_pattern = "{version}" in path_format_str
        if asset_path.exists() and is_versioned_pattern:
            raise FileExistsError(f"Asset path for new version {asset_path} already exists.")

        # Done. Write to disk.
        asset_path.parent.mkdir(parents=True, exist_ok=True)
        with open(asset_path, mode="wb") as f:
            f.write(data)
        return next_uri

    async def list_assets(
        self,
        asset_type: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> List[AssetUri]:
        """
        List assets in the store, optionally filtered by asset type and
        with pagination. For versioned stores, this lists the latest
        version of each asset.
        """
        latest_asset_versions: Dict[Tuple[str, str], str] = {}

        for path_obj in self._base_dir.rglob("*"):
            parsed_uri = self._path_to_uri(path_obj)
            if parsed_uri:
                if asset_type is not None and parsed_uri.asset_type != asset_type:
                    continue

                key = (parsed_uri.asset_type, parsed_uri.asset_id)
                current_version_str = cast(str, parsed_uri.version)  # Is "1" or numeric string

                if key not in latest_asset_versions or int(current_version_str) > int(
                    latest_asset_versions[key]
                ):
                    latest_asset_versions[key] = current_version_str

        result_uris: List[AssetUri] = [
            AssetUri.build(
                asset_type=atype, asset_id=aid, version=vstr
            )  # Params not included in list_assets results
            for (atype, aid), vstr in latest_asset_versions.items()
        ]
        result_uris.sort(key=lambda u: (u.asset_type, u.asset_id, int(cast(str, u.version))))

        start = offset if offset is not None else 0
        end = start + limit if limit is not None else len(result_uris)
        return result_uris[start:end]

    async def count_assets(self, asset_type: Optional[str] = None) -> int:
        """
        Counts assets in the store, optionally filtered by asset type.
        """
        unique_assets: set[Tuple[str, str]] = set()

        for path_obj in self._base_dir.rglob("*"):
            parsed_uri = self._path_to_uri(path_obj)
            if parsed_uri:
                if asset_type is not None and parsed_uri.asset_type != asset_type:
                    continue
                unique_assets.add((parsed_uri.asset_type, parsed_uri.asset_id))

        return len(unique_assets)

    async def list_versions(self, uri: AssetUri) -> List[AssetUri]:
        """
        Lists available version identifiers for a specific asset URI.
        Args:
            uri: The URI of the asset (version component is ignored, params preserved).
        Returns:
            A list of AssetUri objects, sorted by version in ascending order.
        """
        if uri.asset_id is None:
            raise ValueError(f"Asset ID must be specified for listing versions: {uri}")

        path_format_str = self._get_path_format_for_uri_type(uri.asset_type)
        is_versioned_pattern = "{version}" in path_format_str

        if not is_versioned_pattern:
            # Check existence of the single unversioned file
            # Conceptual version is "1", params from input URI are preserved
            path_check_uri = AssetUri.build(
                asset_type=uri.asset_type,
                asset_id=uri.asset_id,
                version="1",
                params=uri.params,
            )
            path_to_asset = self._uri_to_path(path_check_uri)
            path_to_asset = _resolve_case_insensitive(path_to_asset)
            if path_to_asset.is_file():
                return [path_check_uri]  # Returns URI with version "1" and original params
            return []

        found_versions_strs: List[str] = []
        for path_obj in self._base_dir.rglob("*"):
            parsed_uri = self._path_to_uri(path_obj)  # This parsed_uri does not have params
            if (
                parsed_uri
                and parsed_uri.asset_type == uri.asset_type
                and parsed_uri.asset_id == uri.asset_id
            ):
                # Version from path is guaranteed numeric string by _path_to_uri for versioned patterns
                found_versions_strs.append(cast(str, parsed_uri.version))

        if not found_versions_strs:
            return []
        sorted_unique_versions = sorted(list(set(found_versions_strs)), key=int)

        return [
            AssetUri.build(
                asset_type=uri.asset_type,
                asset_id=uri.asset_id,
                version=v_str,
                params=uri.params,
            )
            for v_str in sorted_unique_versions
        ]

    async def is_empty(self, asset_type: Optional[str] = None) -> bool:
        """
        Checks if the store contains any assets, optionally filtered by asset
        type.
        """
        # Reuses list_assets which iterates files.
        # Limit=1 makes it stop after finding the first asset.
        assets = await self.list_assets(asset_type=asset_type, limit=1)
        return not bool(assets)



# ***************************************************************************
# *   Copyright (c) 2023 edi <edi271@a1.net>                                *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the TechDraw AddOffsetVertex Task Dialog."""

__title__ = "TechDrawTools.TasAddOffsetVertex"
__author__ = "edi"
__url__ = "https://www.freecad.org"
__version__ = "00.01"
__date__ = "2023/12/04"

import FreeCAD as App
import FreeCADGui as Gui
import TechDraw

from functools import partial

import os

translate = App.Qt.translate

class TaskAddOffsetVertex():
    '''Provides the TechDraw AddOffsetVertex Task Dialog.'''
    def __init__(self,view,vertex):
        self._uiPath = App.getHomePath()
        self._uiPath = os.path.join(self._uiPath, "Mod/TechDraw/TechDrawTools/Gui/TaskAddOffsetVertex.ui")
        self.form = Gui.PySideUic.loadUi(self._uiPath)
        self.form.setWindowTitle(translate("TechDraw_AddOffsetVertex", "Add offset vertex"))
        self.view = view
        self.vertex = vertex

        App.setActiveTransaction("Add offset vertex")

    def accept(self):
        '''slot: OK pressed'''
        point = self.vertex.Point   # this is unscaled and inverted, but is also rotated.
        # unrotate point. Note that since this is already unscaled, we need to set the
        # third parameter to False to avoid an extra descaling.
        point = TechDraw.makeCanonicalPoint(self.view, point, False);
        xOffset = self.form.dSpinBoxX.value()
        yOffset = self.form.dSpinBoxY.value()
        offset = App.Vector(xOffset,yOffset,0)  # the offset is applied to the canonical
                                                # point.  it is an unscaled, unrotated,
                                                # uninverted relative value.
        self.view.makeCosmeticVertex(point+offset)
        Gui.Control.closeDialog()
        App.closeActiveTransaction()

    def reject(self):
        App.closeActiveTransaction(True)
        return True



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2019 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import json
import math

import ifcopenshell
from ifcopenshell import guid

import FreeCAD

# import Draft

from draftutils import params


def getObjectsOfIfcType(objects, ifcType):
    results = []
    for object in objects:
        if hasattr(object, "IfcType"):
            if object.IfcType == ifcType:
                results.append(object)
    return results


def writeUnits(ifcfile, unit="metre"):
    """adds additional units settings to the given ifc file if needed"""
    # so far, only metre or foot possible (which is all revit knows anyway)

    if unit == "foot":
        d1 = ifcfile.createIfcDimensionalExponents(1, 0, 0, 0, 0, 0, 0)
        d2 = ifcfile.createIfcMeasureWithUnit(ifcfile.createIfcRatioMeasure(0.3048), ifcfile[13])
        d3 = ifcfile.createIfcConversionBasedUnit(d1, "LENGTHUNIT", "FOOT", d2)
        d4 = ifcfile.createIfcDimensionalExponents(2, 0, 0, 0, 0, 0, 0)
        d5 = ifcfile.createIfcMeasureWithUnit(
            ifcfile.createIfcRatioMeasure(0.09290304000000001), ifcfile[14]
        )
        d6 = ifcfile.createIfcConversionBasedUnit(d4, "AREAUNIT", "SQUARE FOOT", d5)
        d7 = ifcfile.createIfcDimensionalExponents(3, 0, 0, 0, 0, 0, 0)
        d8 = ifcfile.createIfcMeasureWithUnit(
            ifcfile.createIfcRatioMeasure(0.028316846592), ifcfile[15]
        )
        d9 = ifcfile.createIfcConversionBasedUnit(d7, "VOLUMEUNIT", "CUBIC FOOT", d8)
        ifcfile.createIfcUnitAssignment((d3, d6, d9, ifcfile[18]))
    else:  # default = metre, no need to add anything
        ifcfile.createIfcUnitAssignment((ifcfile[13], ifcfile[14], ifcfile[15], ifcfile[18]))
    return ifcfile


def writeQuantities(ifcfile, obj, product, history, scale):
    "append quantities to the given object"

    if hasattr(obj, "IfcData"):
        quantities = []
        if (
            ("ExportHeight" in obj.IfcData)
            and obj.IfcData["ExportHeight"]
            and hasattr(obj, "Height")
        ):
            quantities.append(
                ifcfile.createIfcQuantityLength("Height", None, None, obj.Height.Value * scale)
            )
        if ("ExportWidth" in obj.IfcData) and obj.IfcData["ExportWidth"] and hasattr(obj, "Width"):
            quantities.append(
                ifcfile.createIfcQuantityLength("Width", None, None, obj.Width.Value * scale)
            )
        if (
            ("ExportLength" in obj.IfcData)
            and obj.IfcData["ExportLength"]
            and hasattr(obj, "Length")
        ):
            quantities.append(
                ifcfile.createIfcQuantityLength("Length", None, None, obj.Length.Value * scale)
            )
        if (
            ("ExportHorizontalArea" in obj.IfcData)
            and obj.IfcData["ExportHorizontalArea"]
            and hasattr(obj, "HorizontalArea")
        ):
            quantities.append(
                ifcfile.createIfcQuantityArea(
                    "HorizontalArea", None, None, obj.HorizontalArea.Value * (scale**2)
                )
            )
        if (
            ("ExportVerticalArea" in obj.IfcData)
            and obj.IfcData["ExportVerticalArea"]
            and hasattr(obj, "VerticalArea")
        ):
            quantities.append(
                ifcfile.createIfcQuantityArea(
                    "VerticalArea", None, None, obj.VerticalArea.Value * (scale**2)
                )
            )
        if (
            ("ExportVolume" in obj.IfcData)
            and obj.IfcData["ExportVolume"]
            and obj.isDerivedFrom("Part::Feature")
        ):
            quantities.append(
                ifcfile.createIfcQuantityVolume("Volume", None, None, obj.Shape.Volume * (scale**3))
            )
        if quantities:
            eltq = ifcfile.createIfcElementQuantity(
                ifcopenshell.guid.new(), history, "ElementQuantities", None, "FreeCAD", quantities
            )
            ifcfile.createIfcRelDefinesByProperties(
                ifcopenshell.guid.new(), history, None, None, [product], eltq
            )


class SIUnitCreator:
    def __init__(self, file, text, type):
        self.prefixes = [
            "EXA",
            "PETA",
            "TERA",
            "GIGA",
            "MEGA",
            "KILO",
            "HECTO",
            "DECA",
            "DECI",
            "CENTI",
            "MILLI",
            "MICRO",
            "NANO",
            "PICO",
            "FEMTO",
            "ATTO",
        ]
        self.unitNames = [
            "AMPERE",
            "BECQUEREL",
            "CANDELA",
            "COULOMB",
            "CUBIC_METRE",
            "DEGREE CELSIUS",
            "FARAD",
            "GRAM",
            "GRAY",
            "HENRY",
            "HERTZ",
            "JOULE",
            "KELVIN",
            "LUMEN",
            "LUX",
            "MOLE",
            "NEWTON",
            "OHM",
            "PASCAL",
            "RADIAN",
            "SECOND",
            "SIEMENS",
            "SIEVERT",
            "SQUARE METRE",
            "METRE",
            "STERADIAN",
            "TESLA",
            "VOLT",
            "WATT",
            "WEBER",
        ]
        self.text = text
        self.SIUnit = file.createIfcSIUnit(None, type, self.getSIPrefix(), self.getSIUnitName())

    def getSIPrefix(self):
        for prefix in self.prefixes:
            if prefix in self.text.upper():
                return prefix
        return None

    def getSIUnitName(self):
        for unitName in self.unitNames:
            if unitName in self.text.upper():
                return unitName
        return None


class ContextCreator:
    def __init__(self, file, objects):
        self.file = file
        self.objects = objects
        self.project_object = self.getProjectObject()
        self.project_data = self.getProjectObjectData()
        self.model_context = self.createGeometricRepresentationContext()
        self.model_view_subcontext = self.createGeometricRepresentationSubContext()
        self.target_crs = self.createTargetCRS()
        self.map_conversion = self.createMapConversion()
        self.project = self.createProject()

    def createGeometricRepresentationContext(self):
        return self.file.createIfcGeometricRepresentationContext(
            None,
            "Model",
            3,
            1.0e-05,
            self.file.by_type("IfcAxis2Placement3D")[0],
            self.createTrueNorth(),
        )

    def createGeometricRepresentationSubContext(self):
        return self.file.createIfcGeometricRepresentationSubContext(
            "Body", "Model", None, None, None, None, self.model_context, None, "MODEL_VIEW", None
        )

    def createTargetCRS(self):
        try:
            SIUnit = SIUnitCreator(self.file, self.project_data["map_unit"], "LENGTHUNIT")
            return self.file.createIfcProjectedCRS(
                self.project_data["name"],
                self.project_data["description"],
                self.project_data["geodetic_datum"],
                self.project_data["vertical_datum"],
                self.project_data["map_projection"],
                self.project_data["map_zone"],
                SIUnit.SIUnit,
            )
        except Exception:
            return None

    def createMapConversion(self):
        try:
            return self.file.createIfcMapConversion(
                self.model_context,
                self.target_crs,
                float(self.project_data["eastings"]),
                float(self.project_data["northings"]),
                float(self.project_data["orthogonal_height"]),
                self.calculateXAxisAbscissa(),
                self.calculateXAxisOrdinate(),
                float(self.project_data["scale"]),
            )
        except Exception:
            return None

    def createTrueNorth(self):
        return self.file.createIfcDirection(
            (self.calculateXAxisAbscissa(), self.calculateXAxisOrdinate())
        )

    def calculateXAxisAbscissa(self):
        if "true_north" in self.project_data:
            return math.cos(math.radians(float(self.project_data["true_north"]) + 90))
        return 0.0

    def calculateXAxisOrdinate(self):
        if "true_north" in self.project_data:
            return math.sin(math.radians(float(self.project_data["true_north"]) + 90))
        return 1.0

    def createProject(self):
        if not self.project_object:
            return self.createAutomaticProject()
        return self.createCustomProject()

    def createAutomaticProject(self):
        return self.file.createIfcProject(
            self.getProjectGUID(),
            self.file.by_type("IfcOwnerHistory")[0],
            FreeCAD.ActiveDocument.Name,
            None,
            None,
            None,
            None,
            [self.model_context],
            self.file.by_type("IfcUnitAssignment")[0],
        )

    def createCustomProject(self):
        return self.file.createIfcProject(
            self.getProjectGUID(),
            self.file.by_type("IfcOwnerHistory")[0],
            self.project_object.Label,
            self.project_object.Description,
            self.project_object.ObjectType,
            self.project_object.LongName,
            self.project_object.Phase,
            [self.model_context],
            self.file.by_type("IfcUnitAssignment")[0],
        )

    def getProjectGUID(self):
        # TODO: Do not generate a new one each time, but at least this one
        # conforms to the community consensus on how a GUID is generated.
        return ifcopenshell.guid.new()

    def getProjectObject(self):
        try:
            return getObjectsOfIfcType(self.objects, "Project")[0]
        except Exception:
            return None

    def getProjectObjectData(self):
        if not self.project_object:
            return {}
        return json.loads(self.project_object.IfcData["complex_attributes"])[
            "RepresentationContexts"
        ]


class recycler:
    "the compression engine - a mechanism to reuse ifc entities if needed"

    # this object has some methods identical to corresponding ifcopenshell methods,
    # but it checks if a similar entity already exists before creating a new one
    # to compress a new type, just add the necessary method here

    def __init__(self, ifcfile, template=True):

        self.ifcfile = ifcfile
        self.compress = params.get_param_arch("ifcCompress")
        self.mergeProfiles = params.get_param_arch("ifcMergeProfiles")
        self.cartesianpoints = {}
        self.directions = {}
        self.axis2placement3ds = {}
        if template:  # we are using the default template from exportIFC.py
            self.cartesianpoints = {(0, 0, 0): self.ifcfile[8]}  # from template
            self.directions = {
                (1, 0, 0): self.ifcfile[6],
                (0, 0, 1): self.ifcfile[7],
                (0, 1, 0): self.ifcfile[10],
            }  # from template
            self.axis2placement3ds = {
                "(0.0, 0.0, 0.0)(0.0, 0.0, 1.0)(1.0, 0.0, 0.0)": self.ifcfile[9]
            }  # from template
        self.polylines = {}
        self.polyloops = {}
        self.propertysinglevalues = {}
        self.axis2placement2ds = {}
        self.localplacements = {}
        self.rgbs = {}
        self.ssrenderings = {}
        self.sstyles = {}
        self.transformationoperators = {}
        self.psas = {}
        self.spared = 0
        self.profiledefs = {}

    def createIfcCartesianPoint(self, points):
        if self.compress and points in self.cartesianpoints:
            self.spared += 1
            return self.cartesianpoints[points]
        else:
            c = self.ifcfile.createIfcCartesianPoint(points)
            if self.compress:
                self.cartesianpoints[points] = c
            return c

    def createIfcDirection(self, points):
        if self.compress and points in self.directions:
            self.spared += 1
            return self.directions[points]
        else:
            c = self.ifcfile.createIfcDirection(points)
            if self.compress:
                self.directions[points] = c
            return c

    def createIfcPolyline(self, points):
        key = "".join([str(p.Coordinates) for p in points])
        if self.compress and key in self.polylines:
            self.spared += 1
            return self.polylines[key]
        else:
            c = self.ifcfile.createIfcPolyline(points)
            if self.compress:
                self.polylines[key] = c
            return c

    def createIfcPolyLoop(self, points):
        key = "".join([str(p.Coordinates) for p in points])
        if self.compress and key in self.polyloops:
            self.spared += 1
            return self.polyloops[key]
        else:
            c = self.ifcfile.createIfcPolyLoop(points)
            if self.compress:
                self.polyloops[key] = c
            return c

    def createIfcPropertySingleValue(self, name, ptype, pvalue):
        key = str(name) + str(ptype) + str(pvalue)
        if self.compress and key in self.propertysinglevalues:
            self.spared += 1
            return self.propertysinglevalues[key]
        else:
            if (
                isinstance(pvalue, float) and pvalue < 0.000000001
            ):  # remove the exp notation that some bim apps hate
                pvalue = 0
            c = self.ifcfile.createIfcPropertySingleValue(
                name, None, self.ifcfile.create_entity(ptype, pvalue), None
            )
            if self.compress:
                self.propertysinglevalues[key] = c
            return c

    def createIfcAxis2Placement3D(self, p1=None, p2=None, p3=None):
        if not p1:
            p1 = self.createIfcCartesianPoint((0.0, 0.0, 0.0))
            p2 = self.createIfcDirection((0.0, 0.0, 1.0))
            p3 = self.createIfcDirection((1.0, 0.0, 0.0))
        if p2:
            tp2 = str(p2.DirectionRatios)
        else:
            tp2 = "None"
        if p3:
            tp3 = str(p3.DirectionRatios)
        else:
            tp3 = "None"
        key = str(p1.Coordinates) + tp2 + tp3
        if self.compress and key in self.axis2placement3ds:
            self.spared += 1
            return self.axis2placement3ds[key]
        else:
            c = self.ifcfile.createIfcAxis2Placement3D(p1, p2, p3)
            if self.compress:
                self.axis2placement3ds[key] = c
            return c

    def createIfcAxis2Placement2D(self, p1, p2):
        key = str(p1.Coordinates) + str(p2.DirectionRatios)
        if self.compress and key in self.axis2placement2ds:
            self.spared += 1
            return self.axis2placement2ds[key]
        else:
            c = self.ifcfile.createIfcAxis2Placement2D(p1, p2)
            if self.compress:
                self.axis2placement2ds[key] = c
            return c

    def createIfcLocalPlacement(self, gpl=None):
        if not gpl:
            gpl = self.createIfcAxis2Placement3D()
        key = (
            str(gpl.Location.Coordinates)
            + str(gpl.Axis.DirectionRatios)
            + str(gpl.RefDirection.DirectionRatios)
        )
        if self.compress and key in self.localplacements:
            self.spared += 1
            return self.localplacements[key]
        else:
            c = self.ifcfile.createIfcLocalPlacement(None, gpl)
            if self.compress:
                self.localplacements[key] = c
            return c

    def createIfcColourRgb(self, r, g, b):
        key = (r, g, b)
        if self.compress and key in self.rgbs:
            self.spared += 1
            return self.rgbs[key]
        else:
            c = self.ifcfile.createIfcColourRgb(None, r, g, b)
            if self.compress:
                self.rgbs[key] = c
            return c

    def createIfcSurfaceStyleRendering(self, col, alpha=1):
        key = (col.Red, col.Green, col.Blue, alpha)
        if self.compress and key in self.ssrenderings:
            self.spared += 1
            return self.ssrenderings[key]
        else:
            if alpha == 1:
                alpha = None
            c = self.ifcfile.createIfcSurfaceStyleRendering(
                col, alpha, None, None, None, None, None, None, "FLAT"
            )
            if self.compress:
                self.ssrenderings[key] = c
            return c

    def createIfcCartesianTransformationOperator3D(self, axis1, axis2, origin, scale, axis3):
        key = (
            str(axis1.DirectionRatios)
            + str(axis2.DirectionRatios)
            + str(origin.Coordinates)
            + str(scale)
            + str(axis3.DirectionRatios)
        )
        if self.compress and key in self.transformationoperators:
            self.spared += 1
            return self.transformationoperators[key]
        else:
            c = self.ifcfile.createIfcCartesianTransformationOperator3D(
                axis1, axis2, origin, scale, axis3
            )
            if self.compress:
                self.transformationoperators[key] = c
            return c

    def createIfcSurfaceStyle(self, name, r, g, b, a=1):
        if name:
            key = name + str((r, g, b))
        else:
            key = str((r, g, b))
        if self.compress and key in self.sstyles:
            self.spared += 1
            return self.sstyles[key]
        else:
            col = self.createIfcColourRgb(r, g, b)
            ssr = self.createIfcSurfaceStyleRendering(col, a)
            c = self.ifcfile.createIfcSurfaceStyle(name, "BOTH", [ssr])
            if self.compress:
                self.sstyles[key] = c
            return c

    def createIfcPresentationStyleAssignment(self, name, r, g, b, a=1, ifc4=False):
        if name:
            key = name + str((r, g, b, a))
        else:
            key = str((r, g, b, a))
        if self.compress and key in self.psas:
            self.spared += 1
            return self.psas[key]
        else:
            iss = self.createIfcSurfaceStyle(name, r, g, b, a)
            if ifc4:
                c = iss
            else:
                c = self.ifcfile.createIfcPresentationStyleAssignment([iss])
            if self.compress:
                self.psas[key] = c
            return c

    def createIfcRectangleProfileDef(self, name, mode, pt, b, h):
        key = "RECT" + str(name) + str(mode) + str(pt) + str(b) + str(h)
        if self.compress and self.mergeProfiles and key in self.profiledefs:
            return self.profiledefs[key]
        else:
            c = self.ifcfile.createIfcRectangleProfileDef(name, mode, pt, b, h)
            if self.compress and self.mergeProfiles:
                self.profiledefs[key] = c
            return c

    def createIfcCircleProfileDef(self, name, mode, pt, r):
        key = "CIRC" + str(name) + str(mode) + str(pt) + str(r)
        if self.compress and self.mergeProfiles and key in self.profiledefs:
            return self.profiledefs[key]
        else:
            c = self.ifcfile.createIfcCircleProfileDef(name, mode, pt, r)
            if self.compress and self.mergeProfiles:
                self.profiledefs[key] = c
            return c

    def createIfcEllipseProfileDef(self, name, mode, pt, majr, minr):
        key = "ELLI" + str(name) + str(mode) + str(pt) + str(majr) + str(minr)
        if self.compress and self.mergeProfiles and key in self.profiledefs:
            return self.profiledefs[key]
        else:
            c = self.ifcfile.createIfcEllipseProfileDef(name, mode, pt, majr, minr)
            if self.compress and self.mergeProfiles:
                self.profiledefs[key] = c
            return c



# SPDX-License-Identifier: LGPL-2.1-or-later

import FreeCAD as App
import Part

import unittest


class BRepTests(unittest.TestCase):

    def testProject(self):
        """
        This is a unit test for PR #13507
        """
        num = 18
        alt = [0, 1] * num
        pts = [App.Vector(i, alt[i], 0) for i in range(num)]

        bsc = Part.BSplineCurve()
        bsc.buildFromPoles(pts, False, 1)
        edge = bsc.toShape()

        rts = Part.RectangularTrimmedSurface(Part.Plane(), -50, 50, -50, 50)
        plane_shape = rts.toShape()

        proj = plane_shape.project([edge])
        self.assertFalse(proj.isNull())
        self.assertEqual(len(proj.Edges), 1)

    def testEdgeSplitFace(self):
        coords2d = [(0.5, -0.5), (1.0, -0.5), (1.0, 0.5), (0.5, 0.5)]
        pts2d = [App.Base.Vector2d(u, v) for u, v in coords2d]
        pts2d.append(pts2d[0])

        sphere = Part.Sphere()
        edges = []
        for i in range(1, len(pts2d)):
            ls = Part.Geom2d.Line2dSegment(pts2d[i - 1], pts2d[i])
            edges.append(ls.toShape(sphere))

        split = edges[0].split(0.25)
        new_edges = split.Edges + edges[1:]
        wire = Part.Wire(new_edges)
        face = Part.Face(wire, "Part::FaceMakerSimple")
        self.assertTrue(face.isValid())

    def testEdgeSplitReplace(self):
        cyl = Part.makeCylinder(2, 5)
        e1 = cyl.Edge3
        split = e1.split([1.0, 2.0])
        newcyl = cyl.replaceShape([(e1, split), (cyl.Vertex2, split.Vertex1)])
        self.assertTrue(newcyl.isValid())



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

import addonmanager_freecad_interface as fci

from AddonManagerTest.gui.test_workers_utility import (
    TestWorkersUtility as AddonManagerTestWorkersUtility,
)
from AddonManagerTest.gui.test_workers_startup import (
    TestWorkersStartup as AddonManagerTestWorkersStartup,
)
from AddonManagerTest.gui.test_installer_gui import (
    TestInstallerGui as AddonManagerTestInstallerGui,
)
from AddonManagerTest.gui.test_installer_gui import (
    TestMacroInstallerGui as AddonManagerTestMacroInstallerGui,
)
from AddonManagerTest.gui.test_update_all_gui import (
    TestUpdateAllGui as AddonManagerTestUpdateAllGui,
)
from AddonManagerTest.gui.test_uninstaller_gui import (
    TestUninstallerGUI as AddonManagerTestUninstallerGUI,
)


class TestListTerminator:
    pass


# Basic usage mostly to get static analyzers to stop complaining about unused imports
loaded_gui_tests = [
    AddonManagerTestWorkersUtility,
    AddonManagerTestWorkersStartup,
    AddonManagerTestInstallerGui,
    AddonManagerTestMacroInstallerGui,
    AddonManagerTestUpdateAllGui,
    AddonManagerTestUninstallerGUI,
    TestListTerminator,  # Needed to prevent the last test from running twice
]
for test in loaded_gui_tests:
    fci.Console.PrintLog(f"Loaded tests from {test.__name__}\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2013 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Auxiliary functions for the unit tests of the Draft Workbench."""

## @package auxiliary
# \ingroup drafttests
# \brief Auxiliary functions for the unit tests of the Draft Workbench.

## \addtogroup drafttests
# @{

import traceback

from draftutils.messages import _msg


def draw_header():
    """Draw a header for the tests."""
    _msg("")
    _msg(78 * "-")


def import_test(module):
    """Try importing a module."""
    _msg("  Try importing '{}'".format(module))
    try:
        imported = __import__("{}".format(module))
    except ImportError as exc:
        imported = False
        _msg("  {}".format(exc))
        _msg(traceback.format_exc())
    return imported


def no_gui(module):
    """Print a message that there is no user interface."""
    _msg(
        "  #-----------------------------------------------------#\n"
        "  #    No GUI; cannot test for '{}'\n"
        "  #-----------------------------------------------------#\n"
        "  Automatic PASS".format(module)
    )


def no_test():
    """Print a message that the test is not currently implemented."""
    _msg(
        "  #-----------------------------------------------------#\n"
        "  #    This test is not implemented currently\n"
        "  #-----------------------------------------------------#\n"
        "  Automatic PASS"
    )


def fake_function(p1=None, p2=None, p3=None, p4=None, p5=None):
    """Print a message for a test that doesn't actually exist."""
    _msg("  Arguments to placeholder function")
    _msg("  p1={0}; p2={1}".format(p1, p2))
    _msg("  p3={0}; p4={1}".format(p3, p4))
    _msg("  p5={}".format(p5))
    no_test()
    return True


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2021 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


"""This module contains FreeCAD commands for the Draft workbench"""

import os
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
from draftgeoutils.general import geomType
from draftobjects.base import DraftObject
from draftutils import gui_utils
from draftutils.messages import _err
from draftutils.translate import translate


class Hatch(DraftObject):

    def __init__(self, obj):

        obj.Proxy = self
        self.Type = "Hatch"
        self.setProperties(obj)

    def setProperties(self, obj):

        pl = obj.PropertiesList
        if not "Base" in pl:
            obj.addProperty(
                "App::PropertyLink",
                "Base",
                "Hatch",
                QT_TRANSLATE_NOOP("App::Property", "The base object used by this object"),
                locked=True,
            )
        if not "File" in pl:
            obj.addProperty(
                "App::PropertyFile",
                "File",
                "Hatch",
                QT_TRANSLATE_NOOP("App::Property", "The PAT file used by this object"),
                locked=True,
            )
        if not "Pattern" in pl:
            obj.addProperty(
                "App::PropertyString",
                "Pattern",
                "Hatch",
                QT_TRANSLATE_NOOP("App::Property", "The pattern name used by this object"),
                locked=True,
            )
        if not "Scale" in pl:
            obj.addProperty(
                "App::PropertyFloat",
                "Scale",
                "Hatch",
                QT_TRANSLATE_NOOP("App::Property", "The pattern scale used by this object"),
                locked=True,
            )
        if not "Rotation" in pl:
            obj.addProperty(
                "App::PropertyAngle",
                "Rotation",
                "Hatch",
                QT_TRANSLATE_NOOP("App::Property", "The pattern rotation used by this object"),
                locked=True,
            )
        if not "Translate" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "Translate",
                "Hatch",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "If set to False, hatch is applied as is to the faces, without translation (this might give wrong results for non-XY faces)",
                ),
                locked=True,
            )
            obj.Translate = True

    def onDocumentRestored(self, obj):
        self.setProperties(obj)
        super().onDocumentRestored(obj)
        gui_utils.restore_view_object(
            obj, vp_module="view_hatch", vp_class="ViewProviderDraftHatch"
        )

    def dumps(self):
        return

    def loads(self, state):
        self.Type = "Hatch"

    def execute(self, obj):

        if (
            self.props_changed_placement_only(obj)
            or not obj.Base
            or not obj.File
            or not obj.Pattern
            or not obj.Scale
            or not obj.Base.isDerivedFrom("Part::Feature")
            or not obj.Base.Shape.Faces
        ):
            self.props_changed_clear()
            return

        if obj.File[0] == ".":
            # File path relative to the FreeCAD file directory.
            pat_file = os.path.join(os.path.dirname(obj.Document.FileName), obj.File)
            # We need the absolute path to do some file checks.
            pat_file = os.path.abspath(pat_file)
        else:
            pat_file = obj.File

        # File checks:
        if not os.path.exists(pat_file):
            _err(obj.Label + ": " + translate("draft", "PAT file not found"))
            self.props_changed_clear()
            return
        if not os.path.isfile(pat_file):
            _err(obj.Label + ": " + translate("draft", "Specified PAT file is not a file"))
            self.props_changed_clear()
            return
        if os.path.splitext(pat_file)[1].lower() != ".pat":
            _err(obj.Label + ": " + translate("draft", "Specified file type is not supported"))
            self.props_changed_clear()
            return
        if not obj.Pattern in self.getPatterns(pat_file):
            _err(obj.Label + ": " + translate("draft", "Pattern not found in PAT file"))
            self.props_changed_clear()
            return

        import Part
        import TechDraw

        # In TechDraw edges longer than 9999.9 (ca. 10m) are considered 'crazy'.
        # Lines in hatch patterns are also checked. We need to change a parameter:
        param_grp = App.ParamGet("User parameter:BaseApp/Preferences/Mod/TechDraw/debug")
        if "allowCrazyEdge" not in param_grp.GetBools():
            old_allow_crazy_edge = None
        else:
            old_allow_crazy_edge = param_grp.GetBool("allowCrazyEdge")
        param_grp.SetBool("allowCrazyEdge", True)

        shapes = []
        for face in obj.Base.Shape.Faces:
            if face.findPlane():  # Only planar faces.
                face = face.copy()
                if obj.Translate:
                    mtx = None
                    w = face.normalAt(0, 0)
                    # Try to base a matrix on the first straight edge with
                    # a reasonable length (> 0.001):
                    for e in face.Edges:
                        if geomType(e) == "Line":
                            sta = e.firstVertex().Point
                            end = e.lastVertex().Point
                            u = end.sub(sta)
                            if u.Length > 0.001:
                                u = u.normalize()
                                v = w.cross(u)
                                # fmt: off
                                mtx = App.Matrix(u.x, v.x, w.x, sta.x,
                                                 u.y, v.y, w.y, sta.y,
                                                 u.z, v.z, w.z, sta.z,
                                                 0.0, 0.0, 0.0, 1.0)
                                # fmt: on
                                break
                    # If no suitable straight edge was found use a default matrix:
                    if not mtx:
                        cen = face.CenterOfMass
                        rot = App.Rotation(App.Vector(0, 0, 1), w)
                        mtx = App.Placement(cen, rot).Matrix
                    face = face.transformShape(mtx.inverse()).Faces[0]
                if obj.Rotation.Value:
                    face.rotate(App.Vector(), App.Vector(0, 0, 1), -obj.Rotation)

                shape = TechDraw.makeGeomHatch(face, obj.Scale, obj.Pattern, pat_file)

                if obj.Rotation.Value:
                    shape.rotate(App.Vector(), App.Vector(0, 0, 1), obj.Rotation)
                if obj.Translate:
                    shape = shape.transformShape(mtx)
                shapes.append(shape)

        if old_allow_crazy_edge is None:
            param_grp.RemBool("allowCrazyEdge")
        else:
            param_grp.SetBool("allowCrazyEdge", old_allow_crazy_edge)

        if shapes:
            obj.Shape = Part.makeCompound(shapes)
        self.props_changed_clear()

    def onChanged(self, obj, prop):

        self.props_changed_store(prop)

    def getPatterns(self, filename):
        """returns a list of pattern names found in a PAT file"""
        patterns = []
        if os.path.exists(filename):
            with open(filename) as patfile:
                for line in patfile:
                    if line.startswith("*"):
                        patterns.append(line.split(",")[0][1:])
        return patterns



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix write inpfile step output"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"


def write_step_output(f, ccxwriter):

    f.write("\n{}\n".format(59 * "*"))
    f.write("** Outputs --> frd file\n")
    if (
        ccxwriter.member.geos_beamsection
        or ccxwriter.member.geos_shellthickness
        or ccxwriter.member.geos_fluidsection
    ):
        if ccxwriter.solver_obj.BeamShellResultOutput3D is False:
            f.write("*NODE FILE, OUTPUT=2d\n")
        else:
            f.write("*NODE FILE, OUTPUT=3d\n")
    else:
        f.write("*NODE FILE\n")
    # MPH write out nodal temperatures if thermomechanical
    if ccxwriter.analysis_type == "thermomech":
        if not ccxwriter.member.geos_fluidsection:
            f.write("U, NT\n")
        else:
            f.write("MF, PS\n")
    elif ccxwriter.analysis_type == "electromagnetic":
        f.write("NT\n")
    else:
        f.write("U\n")
    if not ccxwriter.member.geos_fluidsection:
        f.write("*EL FILE\n")
        variables = "S, E"
        if ccxwriter.analysis_type == "thermomech":
            variables += ", HFL"
        if ccxwriter.solver_obj.MaterialNonlinearity == "nonlinear":
            variables += ", PEEQ"
        if ccxwriter.analysis_type == "electromagnetic":
            variables = "HFL"

        f.write(variables + "\n")

        # dat file
        if ccxwriter.member.cons_fixed or ccxwriter.member.cons_displacement:
            f.write("** outputs --> dat file\n")
        if ccxwriter.member.cons_fixed:
            # reaction forces for all Constraint fixed
            f.write("** reaction forces for Constraint fixed\n")
            for femobj in ccxwriter.member.cons_fixed:
                # femobj --> dict, FreeCAD document object is femobj["Object"]
                f.write("*NODE PRINT, NSET={}, TOTALS=ONLY\n".format(femobj["Object"].Name))
                f.write("RF\n")
        if ccxwriter.member.cons_displacement:
            # reaction forces for Constraint displacement constraining translation
            f.write("** reaction forces for Constraint displacement constraining translation\n")
            for femobj in ccxwriter.member.cons_displacement:
                # femobj --> dict, FreeCAD document object is femobj["Object"]
                if (
                    not femobj["Object"].xFree
                    or not femobj["Object"].yFree
                    or not femobj["Object"].zFree
                ):
                    f.write("*NODE PRINT, NSET={}, TOTALS=ONLY\n".format(femobj["Object"].Name))
                    f.write("RF\n")
        if ccxwriter.member.cons_rigidbody:
            # displacements and reaction forces/moments for Constraint rigid body
            f.write("** displacements and reaction forces/moments for Constraint rigid body\n")
            for femobj in ccxwriter.member.cons_rigidbody:
                # femobj --> dict, FreeCAD document object is femobj["Object"]
                f.write("*NODE PRINT, NSET={}_RefNode\n".format(femobj["Object"].Name))
                f.write("U\n")
                f.write("*NODE PRINT, NSET={}_RotNode\n".format(femobj["Object"].Name))
                f.write("U\n")
                if (
                    femobj["Object"].TranslationalModeX != "Free"
                    or femobj["Object"].TranslationalModeY != "Free"
                    or femobj["Object"].TranslationalModeZ != "Free"
                ):
                    f.write(
                        "*NODE PRINT, NSET={}_RefNode, TOTALS=ONLY\n".format(femobj["Object"].Name)
                    )
                    f.write("RF\n")
                if (
                    femobj["Object"].RotationalModeX != "Free"
                    or femobj["Object"].RotationalModeY != "Free"
                    or femobj["Object"].RotationalModeZ != "Free"
                ):
                    f.write(
                        "*NODE PRINT, NSET={}_RotNode, TOTALS=ONLY\n".format(femobj["Object"].Name)
                    )
                    f.write("RF\n")
        if ccxwriter.member.cons_contact:
            # contact forces for all Constraint contact (only available for face-to-face penalty contact)
            f.write("** contact forces for Constraint contact\n")
            for femobj in ccxwriter.member.cons_contact:
                # femobj --> dict, FreeCAD document object is femobj["Object"]
                f.write(
                    "*CONTACT PRINT, MASTER={}, SLAVE={}\n".format(
                        "IND" + femobj["Object"].Name, "DEP" + femobj["Object"].Name
                    )
                )
                f.write("CF, CFN, CFS\n")
        if any(
            vars(ccxwriter.member).get(f"cons_{key}")
            for key in ["fixed", "displacement", "rigidbody", "contact"]
        ):
            f.write("\n")
        f.write(f"*OUTPUT, FREQUENCY={ccxwriter.solver_obj.OutputFrequency}")

        # there is no need to write all integration point results
        # as long as there is no reader for them
        # see https://forum.freecad.org/viewtopic.php?f=18&t=29060
        # f.write("*NODE PRINT , NSET=" + ccxwriter.ccx_nall + "\n")
        # f.write("U \n")
        # f.write("*EL PRINT , ELSET=" + ccxwriter.ccx_eall + "\n")
        # f.write("S \n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to transform sketches into Draft objects."""
## @package draftify
# \ingroup draftfunctions
# \brief Provides functions to transform sketches into Draft objects.

## \addtogroup draftfunctions
# @{

import lazy_loader.lazy_loader as lz

import FreeCAD as App
import draftutils.gui_utils as gui_utils
import draftmake.make_block as make_block
import draftmake.make_wire as make_wire
import draftmake.make_circle as make_circle
import draftmake.make_bspline as make_bspline
import draftmake.make_bezcurve as make_bezcurve
import draftmake.make_arc_3points as make_arc_3points

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")
DraftGeomUtils = lz.LazyLoader("DraftGeomUtils", globals(), "DraftGeomUtils")


def draftify(objectslist, makeblock=False, delete=True):
    """draftify(objectslist,[makeblock],[delete])

    Turn each object of the given list (objectslist can also be a single
    object) into a Draft parametric wire.

    TODO: support more objects

    Parameters
    ----------
    objectslist :

    makeblock : bool
        If makeblock is True, multiple objects will be grouped in a block.

    delete : bool
        If delete = False, old objects are not deleted
    """

    if not isinstance(objectslist, list):
        objectslist = [objectslist]
    newobjlist = []
    for obj in objectslist:
        if hasattr(obj, "Shape"):
            for cluster in Part.sortEdges(obj.Shape.Edges):
                w = Part.Wire(cluster)
                nobj = draftify_shape(w)
                if nobj is None:
                    nobj = App.ActiveDocument.addObject("Part::Feature", obj.Name)
                    nobj.Shape = w
                newobjlist.append(nobj)
                gui_utils.format_object(nobj, obj)
                # sketches are always in wireframe mode. In Draft we don't like that!
                if App.GuiUp:
                    nobj.ViewObject.DisplayMode = "Flat Lines"
            if delete:
                App.ActiveDocument.removeObject(obj.Name)

    if makeblock:
        return make_block.make_block(newobjlist)
    else:
        if len(newobjlist) == 1:
            return newobjlist[0]
        return newobjlist


def draftify_shape(shape):

    nobj = None
    if DraftGeomUtils.hasCurves(shape):
        if len(shape.Edges) == 1:
            edge = shape.Edges[0]
            edge_type = DraftGeomUtils.geomType(edge)
            if edge_type == "Circle":
                if edge.isClosed():
                    nobj = make_circle.make_circle(edge)
                else:
                    first_parameter = edge.FirstParameter
                    last_parameter = edge.LastParameter
                    points = [
                        edge.Curve.value(first_parameter),
                        edge.Curve.value((first_parameter + last_parameter) / 2),
                        edge.Curve.value(last_parameter),
                    ]
                    nobj = make_arc_3points.make_arc_3points(points)
        # TODO: take into consideration trimmed curves and capture the specific
        # type of BSpline and Bezier that can be converted to a draft object.
        # elif edge_type == "BSplineCurve":
        #     knots = [edge.Curve.value(p) for p in edge.Curve.getKnots()]
        #     nobj = make_bspline.make_bspline(knots, closed=edge.isClosed())
        # elif edge_type == "BezierCurve":
        #     nobj = make_bezcurve.make_bezcurve(edge.Curve.getPoles(),
        #                                        closed=edge.isClosed())
    else:
        nobj = make_wire.make_wire(shape)

    return nobj


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2019 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM Diff command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate


class BIM_Diff:
    def GetResources(self):
        return {
            "Pixmap": "BIM_Diff",
            "MenuText": QT_TRANSLATE_NOOP("BIM_Diff", "IFC Diff"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_Diff", "Shows the difference between two IFC-based documents"
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):
        # you need two documents open, containing BIM objects with same IDs
        # make the main doc the active one before running this script!

        # what will be compared: IDs, geometry, materials. Everything else is discarded.
        from PySide import QtGui
        import Draft
        import Part

        MOVE_TOLERANCE = 0.2  # the max allowed move in mm
        VOL_TOLERANCE = 250  # the max allowed volume diff in mm^3

        documents = FreeCAD.listDocuments()

        if len(documents) == 2:
            reply = QtGui.QMessageBox.question(
                None,
                "",
                translate(
                    "BIM",
                    "The current document must be the main one. The other contains newer objects to merge into it. Ensure that only the objects intended for comparison are visible in both documents. Proceed?",
                ),
                QtGui.QMessageBox.Yes | QtGui.QMessageBox.No,
                QtGui.QMessageBox.No,
            )
            if reply == QtGui.QMessageBox.Yes:
                activedoc = FreeCAD.ActiveDocument
                if list(documents.keys())[0] == activedoc.Name:
                    otherdoc = list(documents.values())[1]
                else:
                    otherdoc = list(documents.values())[0]

                # build lists of BIM objects with IFC ID

                objswithoutid = []  # let's try to match these later on

                activedocids = {}  # main document, the original freecad one
                for obj in activedoc.Objects:
                    if hasattr(obj, "IfcData") and obj.ViewObject.Visibility:
                        if "IfcUID" in obj.IfcData:
                            activedocids[obj.IfcData["IfcUID"]] = obj
                        elif obj.isDerivedFrom("Part::Feature"):  # discard BuildingParts
                            objswithoutid.append(obj)

                otherdocids = {}  # other doc to be merged to the main one
                for obj in otherdoc.Objects:
                    if hasattr(obj, "IfcData") and obj.ViewObject.Visibility:
                        if "IfcUID" in obj.IfcData:
                            otherdocids[obj.IfcData["IfcUID"]] = obj

                toselect = []  # objects to select when finished
                additions = []  # objects added
                subtractions = []  # objects subtracted
                modified = []  # objects modified
                moved = []  # objects moved
                matchanged = []  # object is same, but material changed
                matchangedghost = (
                    []
                )  # store shapes of objects whose material has changed to print a blue ghost later on
                renamed = {}  # object label changes
                propertieschanged = {}  # objects whose IFC properties are different

                for id, obj in otherdocids.items():
                    if id in activedocids:
                        # this object already exists
                        mainobj = activedocids[id]
                        if obj.Label != mainobj.Label:
                            # object has a different name
                            renamed[mainobj.Name] = obj.Label
                        if obj.IfcProperties and (obj.IfcProperties != mainobj.IfcProperties):
                            # properties have changed
                            propertieschanged[id] = obj.IfcProperties
                        if hasattr(obj, "Shape") and hasattr(mainobj, "Shape"):
                            v = abs(obj.Shape.Volume - mainobj.Shape.Volume)
                            if v < VOL_TOLERANCE:
                                # identical volume
                                l = (
                                    obj.Shape.BoundBox.Center.sub(mainobj.Shape.BoundBox.Center)
                                ).Length
                                if l < MOVE_TOLERANCE:
                                    # identical position
                                    if (
                                        abs(obj.Shape.BoundBox.XMin - mainobj.Shape.BoundBox.XMin)
                                        < MOVE_TOLERANCE
                                        and abs(
                                            obj.Shape.BoundBox.YMin - mainobj.Shape.BoundBox.YMin
                                        )
                                        < MOVE_TOLERANCE
                                        and abs(
                                            obj.Shape.BoundBox.YMin - mainobj.Shape.BoundBox.YMin
                                        )
                                        < MOVE_TOLERANCE
                                    ):
                                        # same boundbox
                                        if (
                                            hasattr(obj, "Material")
                                            and hasattr(mainobj, "Material")
                                            and (
                                                obj.Material
                                                and mainobj.Material
                                                and (obj.Material.Label == mainobj.Material.Label)
                                            )
                                            or (obj.Material == mainobj.Material)
                                        ):
                                            # same material names
                                            obj.ViewObject.hide()
                                        else:
                                            print(
                                                "Object",
                                                mainobj.Label,
                                                "material has changed",
                                            )
                                            obj.ViewObject.hide()  # we hide these objects since the shape hasn't changed but we keep their shapes
                                            matchangedghost.append(obj.Shape)
                                            matchanged.append(obj)
                                    else:
                                        print(
                                            "Object",
                                            mainobj.Label,
                                            "shape bound box has changed",
                                        )
                                        toselect.append(obj)
                                        modified.append(obj)
                                else:
                                    print(
                                        "Object",
                                        mainobj.Label,
                                        "position has moved by",
                                        l,
                                        "mm",
                                    )
                                    toselect.append(obj)
                                    moved.append(obj)
                            else:
                                print(
                                    "Object",
                                    mainobj.Label,
                                    "shape has changed by",
                                    v,
                                    "mm^3",
                                )
                                toselect.append(obj)
                                modified.append(obj)
                        else:
                            print(
                                "Object",
                                mainobj.Label,
                                "one of the objects has no shape",
                            )
                            toselect.append(obj)
                    else:
                        print("Object", obj.Label, "does not exist yet in main document")
                        toselect.append(obj)
                        additions.append(obj)

                for id, obj in activedocids.items():
                    if not id in otherdocids:
                        if obj.isDerivedFrom("Part::Feature"):  # don't count building parts
                            print("Object", obj.Label, "does not exist anymore in new document")
                            subtractions.append(obj)

                # try to find our objects without ID
                newids = {}
                for obj in objswithoutid:
                    for id, otherobj in otherdocids.items():
                        if not id in activedocids:
                            if abs(otherobj.Shape.Volume - obj.Shape.Volume) < VOL_TOLERANCE:
                                if (
                                    otherobj.Shape.BoundBox.Center.sub(obj.Shape.BoundBox.Center)
                                ).Length < MOVE_TOLERANCE:
                                    if (
                                        abs(obj.Shape.BoundBox.XMin - otherobj.Shape.BoundBox.XMin)
                                        < MOVE_TOLERANCE
                                        and abs(
                                            obj.Shape.BoundBox.YMin - otherobj.Shape.BoundBox.YMin
                                        )
                                        < MOVE_TOLERANCE
                                        and abs(
                                            obj.Shape.BoundBox.YMin - otherobj.Shape.BoundBox.YMin
                                        )
                                        < MOVE_TOLERANCE
                                    ):
                                        # shapes are identical. It's the same object!
                                        newids[obj.Name] = id
                                        break
                    else:
                        print(
                            "Object",
                            obj.Label,
                            "has no ID and was not found in the new document",
                        )
                        subtractions.append(obj)

                matnames = {}  # existing materials
                for obj in activedoc.Objects:
                    if Draft.getType(obj) == "Material":
                        matnames[obj.Label] = obj

                newmats = {}  # new materials
                for obj in otherdoc.Objects:
                    if Draft.getType(obj) == "Material":
                        if not obj.Label in matnames:
                            print("Material", obj.Label, "does not exist in main document")
                            toselect.append(obj)
                            newmats[obj.Label] = obj

                if newmats:
                    group = otherdoc.addObject("App::DocumentObjectGroup", "New_materials")
                    for newmat in newmats.values():
                        group.addObject(newmat)

                if toselect:
                    FreeCAD.setActiveDocument(otherdoc.Name)
                    FreeCAD.ActiveDocument = FreeCADGui.getDocument(otherdoc.Name)
                    FreeCADGui.ActiveDocument = FreeCADGui.getDocument(otherdoc.Name)
                    FreeCADGui.Selection.clearSelection()
                    for obj in toselect:
                        FreeCADGui.Selection.addSelection(obj)

                if additions:
                    shape = Part.makeCompound([a.Shape for a in additions])
                    obj = activedoc.addObject("Part::Feature", "Additions")
                    obj.Shape = shape
                    obj.ViewObject.LineWidth = 5
                    obj.ViewObject.LineColor = (0.0, 1.0, 0.0)
                    obj.ViewObject.ShapeColor = (0.0, 1.0, 0.0)
                    obj.ViewObject.Transparency = 60

                if subtractions:
                    shape = Part.makeCompound([s.Shape for s in subtractions])
                    obj = activedoc.addObject("Part::Feature", "Subtractions")
                    obj.Shape = shape
                    obj.ViewObject.LineWidth = 5
                    obj.ViewObject.LineColor = (1.0, 0.0, 0.0)
                    obj.ViewObject.ShapeColor = (1.0, 0.0, 0.0)
                    obj.ViewObject.Transparency = 60

                if modified:
                    shape = Part.makeCompound([m.Shape for m in modified])
                    obj = activedoc.addObject("Part::Feature", "Modified")
                    obj.Shape = shape
                    obj.ViewObject.LineWidth = 5
                    obj.ViewObject.LineColor = (1.0, 0.5, 0.0)
                    obj.ViewObject.ShapeColor = (1.0, 0.5, 0.0)
                    obj.ViewObject.Transparency = 60

                if moved:
                    shape = Part.makeCompound([m.Shape for m in moved])
                    obj = activedoc.addObject("Part::Feature", "Moved")
                    obj.Shape = shape
                    obj.ViewObject.LineWidth = 5
                    obj.ViewObject.LineColor = (1.0, 1.0, 0.0)
                    obj.ViewObject.ShapeColor = (1.0, 1.0, 0.0)
                    obj.ViewObject.Transparency = 60

                if matchangedghost:
                    shape = Part.makeCompound(matchangedghost)
                    obj = otherdoc.addObject("Part::Feature", "Material_changed")
                    obj.Shape = shape
                    obj.ViewObject.LineWidth = 1
                    obj.ViewObject.LineColor = (0.0, 0.0, 1.0)
                    obj.ViewObject.ShapeColor = (0.0, 0.0, 1.0)
                    obj.ViewObject.Transparency = 90

                if matchanged:
                    reply = QtGui.QMessageBox.question(
                        None,
                        "",
                        str(len(matchanged))
                        + " "
                        + translate(
                            "BIM",
                            "objects still have the same shape but have a different material. Update them in the main document?",
                        ),
                        QtGui.QMessageBox.Yes | QtGui.QMessageBox.No,
                        QtGui.QMessageBox.No,
                    )
                    if reply == QtGui.QMessageBox.Yes:
                        for obj in matchanged:
                            mat = obj.Material
                            if mat:
                                mainobj = activedocids[obj.IfcData["IfcUID"]]
                                if mainobj.Material:
                                    mainmatlabel = mainobj.Material.Label
                                else:
                                    mainmatlabel = "None"
                                if mat.Label in matnames:
                                    # the new material already exists, just change it
                                    print(
                                        "Changing material of",
                                        mainobj.Label,
                                        "from",
                                        mainmatlabel,
                                        "to",
                                        mat.Label,
                                    )
                                    mainobj.Material = matnames[mat.Label]
                                else:
                                    # copy the material over
                                    newmat = activedoc.addObject(
                                        "App::MaterialObjectPython", "Material"
                                    )
                                    newmat.Label = mat.Label
                                    import ArchMaterial

                                    ArchMaterial._ArchMaterial(newmat)
                                    ArchMaterial._ViewProviderArchMaterial(newmat.ViewObject)
                                    newmat.Material = mat.Material
                                    print(
                                        "Changing material of",
                                        mainobj.Label,
                                        "from",
                                        mainmatlabel,
                                        "to",
                                        newmat.Label,
                                    )
                                    mainobj.Material = newmat
                                    matnames[newmat.Label] = newmat

                if newids:
                    reply = QtGui.QMessageBox.question(
                        None,
                        "",
                        str(len(newids))
                        + " "
                        + translate(
                            "BIM",
                            "objects have no IFC ID in the main document, but an identical object with an ID exists in the new document. Transfer these IDs to the original objects?",
                        ),
                        QtGui.QMessageBox.Yes | QtGui.QMessageBox.No,
                        QtGui.QMessageBox.No,
                    )
                    if reply == QtGui.QMessageBox.Yes:
                        for name, id in newids.items():
                            obj = activedoc.getObject(name)
                            if obj:
                                print("Transferring new ID to object", obj.Label)
                                a = obj.IfcData
                                a["IfcUID"] = id
                                obj.IfcData = a

                if renamed:
                    reply = QtGui.QMessageBox.question(
                        None,
                        "",
                        str(len(renamed))
                        + " "
                        + translate("BIM", "objects had their name changed. Rename them?"),
                        QtGui.QMessageBox.Yes | QtGui.QMessageBox.No,
                        QtGui.QMessageBox.No,
                    )
                    if reply == QtGui.QMessageBox.Yes:
                        for name, label in renamed.items():
                            obj = activedoc.getObject(name)
                            if obj:
                                print("Renaming object", obj.Label, "to", label)
                                obj.Label = label

                if propertieschanged:
                    reply = QtGui.QMessageBox.question(
                        None,
                        "",
                        str(len(propertieschanged))
                        + " "
                        + translate("BIM", "objects had their properties changed. Update?"),
                        QtGui.QMessageBox.Yes | QtGui.QMessageBox.No,
                        QtGui.QMessageBox.No,
                    )
                    if reply == QtGui.QMessageBox.Yes:
                        for id, prop in propertieschanged.items():
                            obj = activedocids[id]
                            print("Updating properties of ", obj.Label)
                            obj.IfcProperties = prop

                if moved:
                    reply = QtGui.QMessageBox.question(
                        None,
                        "",
                        str(len(moved))
                        + " "
                        + translate(
                            "BIM",
                            "objects have their location changed. Move them to their new position?",
                        ),
                        QtGui.QMessageBox.Yes | QtGui.QMessageBox.No,
                        QtGui.QMessageBox.No,
                    )
                    if reply == QtGui.QMessageBox.Yes:
                        for obj in moved:
                            mainobj = activedocids[obj.IfcData["IfcUID"]]
                            otherobj = otherdocids[obj.IfcData["IfcUID"]]
                            delta = otherobj.Shape.BoundBox.Center.sub(
                                mainobj.Shape.BoundBox.Center
                            )
                            print("Moving object ", mainobj.Label)
                            Draft.move(mainobj, delta)
                    reply = QtGui.QMessageBox.question(
                        None,
                        "",
                        translate(
                            "BIM",
                            "Colorize the objects that have moved in yellow in the other file (to serve as a diff)?",
                        ),
                        QtGui.QMessageBox.Yes | QtGui.QMessageBox.No,
                        QtGui.QMessageBox.No,
                    )
                    if reply == QtGui.QMessageBox.Yes:
                        for obj in moved:
                            otherobj = otherdocids[obj.IfcData["IfcUID"]]
                            try:
                                otherobj.ViewObject.LineColor = (1.0, 1.0, 0.0)
                                otherobj.ViewObject.ShapeColor = (1.0, 1.0, 0.0)
                                otherobj.ViewObject.Transparency = 60
                            except AttributeError:
                                print(otherobj.Label, "cannot be colorized")

                if modified:
                    reply = QtGui.QMessageBox.question(
                        None,
                        "",
                        translate(
                            "BIM",
                            "Colorize the objects that have been modified in orange in the other file (to serve as a diff)?",
                        ),
                        QtGui.QMessageBox.Yes | QtGui.QMessageBox.No,
                        QtGui.QMessageBox.No,
                    )
                    if reply == QtGui.QMessageBox.Yes:
                        for obj in modified:
                            otherobj = otherdocids[obj.IfcData["IfcUID"]]
                            try:
                                otherobj.ViewObject.LineColor = (1.0, 0.5, 0.0)
                                otherobj.ViewObject.ShapeColor = (1.0, 0.5, 0.0)
                                otherobj.ViewObject.Transparency = 60
                            except AttributeError:
                                print(otherobj.Label, "cannot be colorized")

                if subtractions:
                    reply = QtGui.QMessageBox.question(
                        None,
                        "",
                        str(len(subtractions))
                        + " "
                        + translate(
                            "BIM",
                            "objects do not exist anymore in the new document. Move them to a 'To Delete' group?",
                        ),
                        QtGui.QMessageBox.Yes | QtGui.QMessageBox.No,
                        QtGui.QMessageBox.No,
                    )
                    if reply == QtGui.QMessageBox.Yes:
                        group = activedoc.addObject("App::DocumentObjectGroup", "ToDelete")
                        group.Label = "To Delete"
                        for obj in subtractions:
                            group.addObject(obj)
                    reply = QtGui.QMessageBox.question(
                        None,
                        "",
                        translate(
                            "BIM",
                            "Colorize the objects that have been removed in red in the other file (to serve as a diff)?",
                        ),
                        QtGui.QMessageBox.Yes | QtGui.QMessageBox.No,
                        QtGui.QMessageBox.No,
                    )
                    if reply == QtGui.QMessageBox.Yes:
                        for obj in subtractions:
                            otherobj = otherdoc.addObject("Part::Feature", "Deleted")
                            otherobj.Shape = obj.Shape
                            otherobj.ViewObject.LineColor = (1.0, 0.0, 0.0)
                            otherobj.ViewObject.ShapeColor = (1.0, 0.0, 0.0)
                            otherobj.ViewObject.Transparency = 60

                if additions:
                    reply = QtGui.QMessageBox.question(
                        None,
                        "",
                        translate(
                            "BIM",
                            "Colorize the objects that have been added in green in the other file (to serve as a diff)?",
                        ),
                        QtGui.QMessageBox.Yes | QtGui.QMessageBox.No,
                        QtGui.QMessageBox.No,
                    )
                    if reply == QtGui.QMessageBox.Yes:
                        for obj in additions:
                            otherobj = otherdocids[obj.IfcData["IfcUID"]]
                            try:
                                otherobj.ViewObject.LineColor = (0.0, 1.0, 0.0)
                                otherobj.ViewObject.ShapeColor = (0.0, 1.0, 0.0)
                                otherobj.ViewObject.Transparency = 60
                            except AttributeError:
                                print(otherobj.Label, "cannot be colorized")

        else:
            QtGui.QMessageBox.information(
                None,
                "",
                translate(
                    "BIM",
                    "Two documents are required to be open to run this tool. One which is the main document, and one that contains new objects to compare against the existing one. Make sure only the objects to compare in both documents are visible.",
                ),
            )


FreeCADGui.addCommand("BIM_Diff", BIM_Diff())
